import re
import time
from pyqtgraph.Qt import QtWidgets


def parse_params(text, puzzle):
    """
    Parse a string of the following format to construct references to :class:`~puzzlepiece.param.BaseParam` objects:
    ``[Piece name]:[param name], [Piece name]:[param name], ...``

    :param text: The string to parse.
    :param puzzle: The app's :class:`~puzzlepiece.puzzle.Puzzle`.
    :rtype: list(puzzlepiece.param.BaseParam, )
    """
    args = text.split(", ")
    result = []
    for arg in args:
        piece, name = arg.split(":")
        if name in puzzle.pieces[piece].params:
            result.append(puzzle.pieces[piece].params[name])
        else:
            raise SyntaxError(f"Parameter parse error for {arg}")
    return result


def run(text, puzzle):
    """
    Execute a set of puzzlepiece script commands.

    Commands can be separated by new lines, or a semicolon (``;``) + space.

    If a command needs to *include* a semicolon followed by a space,
    it can be inserted as ``\;`` (backslash character + semicolon + space).

    Comments can be added by starting the line with ``#``.

    The available commands are:
        - ``set:[Piece name]:[param name]:[value]``
        - ``run:[Piece name]:[action name]``
        - ``get:[Piece name]:[param name]``
        - ``sleep:[duration in s]``
        - ``prompt:[text]``
        - ``print:[text]``

    :param text: The string to parse.
    :param puzzle: The app's :class:`~puzzlepiece.puzzle.Puzzle`.
    """
    text = text.replace("\\;", "<!--semicolon-->")
    text = text.replace("; ", "\n")
    instructions = text.split("\n")
    for instruction in instructions:
        if len(instruction) == 0 or instruction[0] == "#":
            continue
        task, *params = instruction.split(":")
        if task == "set":
            piece, param, *value = params
            value = ":".join(value)
            value = value.replace("<!--semicolon-->", ";")
            puzzle.pieces[piece].params[param].set_value(value)
        elif task == "run":
            piece, *action = params
            action = ":".join(action)
            puzzle.pieces[piece].actions[action]()
        elif task == "get":
            piece, *param = params
            param = ":".join(param)
            puzzle.pieces[piece].params[param].get_value()
        elif task == "sleep":
            duration = params[0]
            time.sleep(float(duration))
        elif task == "prompt":
            text = ":".join(params)
            text = format(text, puzzle)
            box = QtWidgets.QMessageBox()
            box.setText(text)
            box.exec()
        elif task == "print":
            text = ":".join(params)
            text = format(text, puzzle)
            print(text)
        else:
            raise SyntaxError("Unknown task in {}".format(instruction))
        puzzle.process_events()


def format(text, puzzle):
    """
    Insert values of :class:`~puzzlepiece.param.BaseParam` objects
    into a string.

    To insert a value, use the following format:
    ``The value is: {[Piece name]:[param name]}``

    If a param has a getter, it will be called.

    One can also apply additional formatting:
    ``The formatted value is {[Piece name]:[param/readout name];[format]}``
    where `format` follows the 'new' standard as described at https://pyformat.info/.
    For example:
    ``The laser power is {laser:power;:.1f}``

    :param text: The string to parse.
    :param puzzle: The app's :class:`~puzzlepiece.puzzle.Puzzle`.
    :rtype: str
    """
    for match in re.findall(r"{[a-zA-Z0-9:; _\-\.]+}", text):
        m = str(match)
        elements = m[1:-1].split(";")
        param = parse_params(elements[0], puzzle)[0]
        if len(elements) == 1:
            if param._format is not None:
                result = param._format.format(param.get_value())
            else:
                result = str(param.get_value())
        elif len(elements) == 2:
            result = ("{" + elements[1] + "}").format(param.get_value())
        text = text.replace(m, result)
    return text



import puzzlepiece as pzp


class Piece(pzp.Piece):
    def define_params(self):
        pzp.param.text(self, "params", "")(None)

        @pzp.param.text(self, "value", "file")
        def set_param(self, value):
            param_name = self.params["params"].get_value()
            if len(param_name):
                params = pzp.parse.parse_params(param_name, self.puzzle)
                for param in params:
                    param.set_value(value)
            return value



from pyqtgraph.Qt import QtWidgets
from functools import wraps
import math

from .puzzle import PretendPuzzle


class Piece(QtWidgets.QGroupBox):
    """
    A single `Piece` object is an unit of automation - an object that is meant to represent a single
    physical instrument (like a laser) or a particular functionality (like a plotter or a parameter scan).

    Pieces can be assembled into a :class:`~puzzlepiece.puzzle.Puzzle` using the Puzzle's
    :func:`~puzzlepiece.puzzle.Puzzle.add_piece` method.

    :param puzzle: The parent :class:`~puzzlepiece.puzzle.Puzzle`.
    :param custom_horizontal: A bool, the custom layout is displayed to the right of the main controls
                              if True.
    """

    def __init__(self, puzzle=None, custom_horizontal=False, *args, **kwargs):
        super().__init__()
        #: Reference to the parent :class:`~puzzlepiece.puzzle.Puzzle`.
        self.puzzle = puzzle or PretendPuzzle()
        #: Boolean flag. See :func:`~puzzlepiece.piece.Piece.call_stop`
        self.stop = False

        #: dict: A dictionary of this Piece's params (see :class:`~puzzlepiece.param.BaseParam`). You can also directly index the Piece object with the param name.
        self.params = {}
        # A reference to the param dictionary for backwards-compatibility
        self.readouts = self.params
        #: dict: A dictionary of this Piece's actions (see :class:`~puzzlepiece.action.Action`)
        self.actions = {}
        self.shortcuts = {}

        if not self.puzzle.debug:
            self.setup()

        self.define_params()
        self.define_readouts()
        self.define_actions()
        self.folder = None

        if custom_horizontal:
            self.layout = QtWidgets.QHBoxLayout()
        else:
            self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)

        control_layout = QtWidgets.QVBoxLayout()
        control_layout.addLayout(self.param_layout())
        control_layout.addLayout(self.action_layout())
        self.layout.addLayout(control_layout)

        custom_layout = self.custom_layout()
        if custom_layout is not None:
            self.layout.addLayout(custom_layout)

        if custom_layout is None or custom_horizontal:
            control_layout.addStretch()

    def param_layout(self, wrap=1):
        """
        Genereates a `QGridLayout` for the params. Override to set a different wrapping.

        :param wrap: the number of columns the params are displayed in .
        :rtype: QtWidgets.QGridLayout
        """
        layout = QtWidgets.QGridLayout()
        visible_params = [key for key in self.params if self.params[key].visible]
        numrows = math.ceil(len(visible_params) / wrap)
        for i, key in enumerate(visible_params):
            layout.addWidget(self.params[key], i % numrows, i // numrows)
        return layout

    def action_layout(self, wrap=2):
        """
        Genereates a `QGridLayout` for the actions. Override to set a different wrapping.

        :param wrap: the number of columns the actions are displayed in.
        :rtype: QtWidgets.QGridLayout
        """
        layout = QtWidgets.QGridLayout()
        visible_actions = [key for key in self.actions if self.actions[key].visible]
        for i, key in enumerate(visible_actions):
            button = QtWidgets.QPushButton(key)
            button.clicked.connect(lambda x=False, _key=key: self.actions[_key]())
            layout.addWidget(button, i // wrap, i % wrap)
        return layout

    def custom_layout(self):
        """
        Override to generate a custom `QLayout` that this Piece will display.

        :rtype: QtWidgets.QLayout
        """
        return None

    def define_params(self):
        """
        Override to define params using decorators from :mod:`puzzlepiece.param`.
        """
        pass

    def define_readouts(self):
        """
        Mostly deprecated.

        Override to define readouts (params with getters). This is no different that defining them in
        :func:`~puzzlepiece.piece.Piece.define_params`, but may be a convenient way to organise the
        definitions within your custom class.
        """
        pass

    def define_actions(self):
        """
        Override to define actions using decorators from :mod:`puzzlepiece.action`.
        """
        pass

    def setup(self):
        """
        Only called if the :class:`~puzzlepiece.puzzle.Puzzle` debug flag is False.
        Override to set up necessary hardware libraries.
        """
        pass

    def open_popup(self, popup, name=None):
        """
        Open a popup window for this Piece. A popup is a :class:`puzzlepiece.piece.Popup`
        object, which is like a Piece but floats in a separate window attached to the main
        :class:`~puzzlepiece.puzzle.Puzzle`. This can be used for handling additional tasks
        that you don't want to clutter the main Piece. See :class:`puzzlepiece.piece.Popup`
        for details on implementing a Popup.

        :param popup: a :class:`puzzlepiece.piece.Popup` _class_ to instantiate
        :param name: text to show as the window title
        :rtype: puzzlepiece.piece.Popup
        """
        # Instantiate the popup
        if isinstance(popup, type):
            popup = popup(self, self.puzzle)
        popup.setStyleSheet("QGroupBox {border:0;}")

        # Make a dialog window for the popup to live in
        dialog = _QDialog(self, popup)
        layout = QtWidgets.QVBoxLayout()
        dialog.setLayout(layout)
        layout.addWidget(popup)
        dialog.setWindowTitle(name or "Popup")

        # Display the dialog
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()
        self.puzzle._close_popups.connect(dialog.accept)

        return popup

    def call_stop(self):
        """
        This method is called by the parent Puzzle when a global stop is called.

        By default, it sets the stop flag to True. Detect the flag in you code to stop processes.

        Alternatively, this can be overriden to support more complex actions.
        """
        self.stop = True

    def handle_close(self, event):
        """
        Only called if the :class:`~puzzlepiece.puzzle.Puzzle` :attr:`~puzzlepiece.puzzle.Puzzle.debug`
        flag is False. Override to disconnect hardware etc when the main window closes.
        """
        pass

    def handle_shortcut(self, event):
        """
        Calls an Action if a keyboard shortcut has been registered.

        :meta private:
        """
        if event.key() in self.shortcuts:
            self.shortcuts[event.key()]()

    def elevate(self):
        """
        If this Piece resides in a :class:`~puzzlepiece.puzzle.Folder`, this method switches the tab
        to make this Piece visible.
        """
        if self.folder is not None:
            self.folder.setCurrentWidget(self)

    def __getitem__(self, name):
        return self.params[name]

    def _ipython_key_completions_(self):
        return self.params.keys()


def ensurer(ensure_function):
    """
    An ensurer is a decorator that can be placed on getters, setters, and actions, and it will run
    ahead of these functions. The intended behaviour is performing checks ahead of running the
    function - for example checking if a laser is connected ahead of trying to set its power.
    This way one ensurer can be written and used in multiple places easily.

    Note that **the method being decorated should raise an exception if the check fails!** This way
    execution will stop if the condition is not met. This is not mandatory though - custom behaviour
    is allowed.

    For example, an ensurer can be defined as a Piece's method (in the main body of the class).::

        @puzzlepiece.piece.ensurer
        def _ensure_connected(self):
            if not self.params['connected'].get_value():
                raise Exception('Laser not connected')

    This can then be used when defining a param (below the param-defining decorator)::

        @puzzlepiece.param.spinbox(self, 'power', 0.)
        @self._ensure_connected
        def power(self, value):
            self.laser.set_power(value)

        @puzzlepiece.param.spinbox(self, 'wavelength', 0.)
        @self._ensure_connected
        def wavelength(self, value):
            self.laser.set_wavelength(value)

    It can also be called directly if preferred, optionally with `capture_exception=True`
    which will return True if the check passes, or False if the check raises an Exception::

        # This should raise an Exception is the check fails
        self._ensure_connected()

        # This will not raise an Exception is the check fails
        if self._ensure_connected(capture_exception=True):
            print("laser is connected!")
    """

    # Decorating a class method with ensure makes it a decorator.
    # Here we create this decorator and return it.
    @wraps(ensure_function)
    def ensure_decorator(self, main_function=None, capture_exception=False):
        if main_function is not None:
            # This means ensure_decorator was used as a decorator, and
            # main_function is the function being decorated. We therefore
            # wrap it with the ensuring functionality and return it
            @wraps(main_function)
            def wrapped_main(self, *args, **kwargs):
                ensure_function(self)
                return main_function(self, *args, **kwargs)

            return wrapped_main
        else:
            # If main_function is None, ensure_decorator has been called
            # directly instead of being used as a decorator, so we
            # just execute ensure_function
            if capture_exception:
                try:
                    ensure_function(self)
                except Exception:
                    return False
                return True
            else:
                ensure_function(self)

    return ensure_decorator


class _QDialog(QtWidgets.QDialog):
    """
    A variant of the QDialog specifically for popups, handles closing them
    with a custom function.
    """

    def __init__(self, parent, popup, *args, **kwargs):
        self.popup = popup
        super().__init__(parent, *args, **kwargs)

    def closeEvent(self, event):
        self.popup.handle_close()
        self.popup.parent_piece.puzzle._close_popups.disconnect(self.accept)
        super().closeEvent(event)


class Popup(Piece):
    """
    A Popup is similar to a Piece, but floats in a separate window attached to the main
    :class:`~puzzlepiece.puzzle.Puzzle`. This can be used for handling additional tasks
    that you don't want to clutter the main Piece. For example you can have a camera
    Piece which can open a Popup to set the camera's region of interest with an interactive
    plot window.

    A Popup can be created and displayed by calling :func:`puzzlepiece.piece.Piece.open_popup`.

    A Popup is attached to a specific Piece and knows it through its
    :attr:`~puzzlepiece.piece.Popup.parent_piece` attribute, but it can also access other
    Pieces through the Puzzle, which it knows through its :attr:`~puzzlepiece.piece.Piece.puzzle`
    attribute.

    A Popup can have params, actions, and custom layouts just like a normal Piece, and are created by
    overriding :func:`~puzzlepiece.piece.Piece.define_params`, :func:`~puzzlepiece.piece.Piece.define_actions`,
    and :func:`~puzzlepiece.piece.Piece.custom_layout` like for a Piece.

    :param puzzle: The parent :class:`~puzzlepiece.puzzle.Puzzle`.
    :param parent_piece: The parent :class:`~puzzlepiece.piece.Piece`.
    :param custom_horizontal: A bool, the custom layout is displayed to the right of the main controls
                              if True.
    """

    def __init__(self, parent_piece, puzzle, custom_horizontal=False, *args, **kwargs):
        self._parent_piece = parent_piece
        super().__init__(puzzle, custom_horizontal, *args, **kwargs)
        self.layout.setContentsMargins(0, 0, 0, 0)

    @property
    def parent_piece(self):
        """
        A reference to this Popup's parent :class:`~puzzlepiece.piece.Piece`,
        the one that created it through :func:`puzzlepiece.piece.Piece.open_popup`.
        """
        return self._parent_piece

    def add_child_params(self, param_names):
        """
        Given a list of param names referring to params of the parent :class:`~puzzlepiece.piece.Piece`,
        add corresponding child params to this Popup.

        This lets you quickly make a Settings popup that adjusts the hidden params of a Piece.

        See :func:`puzzlepiece.param.BaseParam.make_child_param` for details.

        :param param_names: List of the parent_piece's param names to make children from.
        """
        for name in param_names:
            self.params[name] = self.parent_piece.params[name].make_child_param()

    def add_child_actions(self, action_names):
        """
        Given a list of action names referring to actions of the parent :class:`~puzzlepiece.piece.Piece`,
        add corresponding child actions to this Popup.

        This lets you surface additional actions in a Popup without cluttering the main Piece.

        See :func:`puzzlepiece.action.Action.make_child_action` for details.

        :param action_names: List of the parent_piece's action names to make children from.
        """
        for name in action_names:
            self.actions[name] = self.parent_piece.actions[name].make_child_action()

    # TODO: A way to close the Popup from 'within'

    def handle_close(self):
        """
        Called when the Popup is closed. Override to perform actions when the user
        closes this Popup - for example delete related plot elements.

        In contrast to :func:`puzzlepiece.piece.Piece.handle_close`, this is called even
        if the :class:`~puzzlepiece.puzzle.Puzzle` :attr:`~puzzlepiece.puzzle.Puzzle.debug`
        flag is True.
        """
        pass



try:
    from IPython.core.magic import cell_magic, Magics, magics_class
    from IPython import get_ipython
except ModuleNotFoundError:
    raise Exception("This module needs to be imported within an IPython environment.")
import traceback


@magics_class
class CustomMagics(Magics):
    @cell_magic
    def pzp_script(self, line, cell):
        puzzle = line if len(line) else "puzzle"
        self.shell.ex(f'{puzzle}.run("""{cell}""")')

    # The skeleton for this magic comes from https://stackoverflow.com/a/54890975
    @cell_magic
    def safe_run(self, line, cell):
        try:
            self.shell.ex(cell)  # This executes the cell in the current namespace
        except Exception as e:
            try:
                # Check the provided line is callable
                if len(line) and ip.ev(f"callable({line})"):
                    # Save the error to the global namespace and execute the provided handler
                    self.shell.user_ns["error"] = e
                    ip.ev(f"{line}(error)")
            except Exception as ee:
                print("The provided exception handler failed to run:", str(ee))
            # Re-raise the error
            raise e


# Register
ip = get_ipython()
ip.register_magics(CustomMagics)


def format_exception(error):
    return (
        "".join(traceback.format_tb(error.__traceback__))
        + f"\n{type(error).__name__}: {error}"
    )



# Import the automation framework
import puzzlepiece as pzp

# Import the random number generator Piece
from puzzlepiece.pieces import random_number, plotter

# Create a Qt app to contain the whole thing
app = pzp.QApp([])

# Create a Puzzle
puzzle = pzp.Puzzle(app, "Basic example", debug=True)

# Add Pieces to the Puzzle
puzzle.add_piece("random", random_number.Piece, 0, 0)
puzzle.add_piece("plotter", plotter.Piece, 0, 1)

# Show the Puzzle window
puzzle.show()

# Execute the application
# (this enters the Qt loop and will only exit once the window is closed)
app.exec()



from . import parse

from pyqtgraph.Qt import QtWidgets, QtCore
import sys


class Puzzle(QtWidgets.QWidget):
    """
    A container for :class:`puzzlepiece.piece.Piece` objects, meant to be the main QWidget (window)
    of an automation application. It keeps track of the :class:`~puzzlepiece.piece.Piece` objects
    it contains and lets them communicate.

    :param app: A QtApp created to contain this QWidget.
    :param name: A name for the window.
    :param debug: Sets the Puzzle.debug property, if True the app should launch in debug mode and Pieces
        shouldn't communicate with hardware.
    :type debug: bool
    :param bottom_buttons: Whether the bottom buttons of the Puzzle (Tree, Export, STOP) should be shown.
    :type bottom_buttons: bool
    """

    def __init__(
        self, app=None, name="Puzzle", debug=True, bottom_buttons=True, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        # Pieces can handle the debug flag as they wish
        self._debug = debug
        self.app = app or QtWidgets.QApplication.instance()
        self.setWindowTitle(name)
        self._pieces = PieceDict()
        self._globals = Globals()
        # toplevel is used to send keypresses down the QWidget tree,
        # instead of up (which is how they normally propagate).
        # The list stores all the direct children of this QWidget
        self._toplevel = []
        self._threadpool = QtCore.QThreadPool()

        self.wrapper_layout = QtWidgets.QGridLayout()
        self.setLayout(self.wrapper_layout)

        self.layout = QtWidgets.QGridLayout()
        self.wrapper_layout.addLayout(self.layout, 0, 0)

        if bottom_buttons:
            self.wrapper_layout.addLayout(self._button_layout(), 1, 0)

        try:
            # If this doesn't raise a NameError, we're in IPython
            shell = get_ipython()
            # _orig_sys_module_state stores the original IPKernelApp excepthook,
            # irrespective of possible modifications in other cells
            self._old_excepthook = shell._orig_sys_module_state["excepthook"]

            # The following hack allows us to handle exceptions through the Puzzle in IPython.
            # Normally when a cell is executed in an IPython InteractiveShell,
            # sys.excepthook is overwritten with shell.excepthook, and then restored
            # to sys.excepthook after the cell run finishes. Any changes we make to
            # sys.excepthook in here directly will thus be overwritten as soon as the
            # cell that defines the Puzzle finishes running.

            # Instead, we schedule set_excepthook on a QTimer, meaning that it will
            # execute in the Qt loop rather than in a cell, so it can modify
            # sys.excepthook without risk of the changes being immediately overwritten,

            # For bonus points, we could set _old_excepthook to shell.excepthook,
            # which would result in all tracebacks appearing in the Notebook rather
            # than the console, but I think that is not desireable.
            def set_excepthook():
                # Make sure we're out of the cell execution context
                if (
                    set_excepthook.counter < 100
                    and sys.excepthook is not self._old_excepthook
                ):
                    # if not, we wait a little bit more
                    set_excepthook.counter += 1
                    QtCore.QTimer.singleShot(500, set_excepthook)
                else:
                    sys.excepthook = self._excepthook

            set_excepthook.counter = 0

            QtCore.QTimer.singleShot(0, set_excepthook)
        except NameError:
            # In normal Python (not IPython) this is comparatively easy.
            # We use the original system hook here instead of sys.excepthook
            # to avoid unexpected behaviour if multiple things try to override
            # the hook in various ways.
            # If you need to implement custom exception handling, please assign
            # a value to your Puzzle's ``custom_excepthook`` method.
            self._old_excepthook = sys.__excepthook__
            sys.excepthook = self._excepthook

    @property
    def pieces(self):
        """
        A :class:`~puzzlepiece.puzzle.PieceDict`, effectively a dictionary of
        :class:`~puzzlepiece.piece.Piece` objects. Can be used to access Pieces from within other Pieces.

        You can also directly index the Puzzle object with the :class:`~puzzlepiece.piece.Piece` name,
        or even with a :class:`~puzzlepiece.piece.Piece` and a :class:`~puzzlepiece.param.BaseParam`::

            # These two are equivalent
            puzzle.pieces["piece_name"]
            puzzle["piece_name"]

            # These three are equivalent
            puzzle.pieces["piece_name"].params["piece_name"]
            puzzle["piece_name"]["param_name"]
            puzzle["piece_name:param_name"]

        The valid keys for indexing a Puzzle object are available when autocompleting the key in IPython.
        """
        return self._pieces

    @property
    def globals(self):
        """
        A dictionary, can be used for API modules that need to be shared by multiple Pieces.
        """
        return self._globals

    @property
    def debug(self):
        """
        A `bool` flag. Pieces should act in debug mode if `True`.
        """
        return self._debug

    # Adding elements

    def add_piece(self, name, piece, row, column, rowspan=1, colspan=1):
        """
        Adds a :class:`~puzzlepiece.piece.Piece` to the grid layout, and registers it with the Puzzle.

        :param name: Identifying string for the Piece.
        :param piece: A :class:`~puzzlepiece.piece.Piece` object or a class defining one (which will
          be automatically instantiated).
        :param row: Row index for the grid layout.
        :param column: Column index for the grid layout.
        :param rowspan: Height in rows.
        :param column: Width in columns.
        :rtype: puzzlepiece.piece.Piece
        """
        if isinstance(piece, type):
            piece = piece(self)
        self.layout.addWidget(piece, row, column, rowspan, colspan)
        self._toplevel.append(piece)
        self.register_piece(name, piece)

        return piece

    def replace_piece(self, name, new_piece):
        """
        Replace a named :class:`~puzzlepiece.piece.Piece` with a new one. Can be
        combined with ``importlib.reload`` to do live development on Pieces.

        This method is **experimental** and can sometimes fail. It's useful for development,
        but shouldn't really be used in production applications.

        :param name: Name of the Piece to be replaced.
        :param piece: A :class:`~puzzlepiece.piece.Piece` object or a class defining one (which will
          be automatically instantiated).
        """
        old_piece = self.pieces[name]
        if isinstance(new_piece, type):
            new_piece = new_piece(self)

        if old_piece in self._toplevel:
            self.layout.replaceWidget(
                old_piece,
                new_piece,
                options=QtCore.Qt.FindChildOption.FindDirectChildrenOnly,
            )
            new_piece.setTitle(name)
            self._toplevel.remove(old_piece)
            self._toplevel.append(new_piece)
        else:
            for widget in self._toplevel:
                if isinstance(widget, Folder):
                    widget._replace_piece(name, old_piece, new_piece)

        self._pieces._replace_item(name, new_piece)
        old_piece.handle_close(None)
        # old_piece.deleteLater()
        old_piece.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose, True)
        old_piece.close()

    def add_folder(self, row, column, rowspan=1, colspan=1):
        """
        Adds a tabbed :class:`~puzzlepiece.puzzle.Folder` to the grid layout, and returns it.

        :param row: Row index for the grid layout.
        :param column: Column index for the grid layout.
        :param rowspan: Height in rows.
        :param column: Width in columns.

        :rtype: puzzlepiece.puzzle.Folder
        """
        folder = Folder(self)
        self.layout.addWidget(folder, row, column, rowspan, colspan)
        self._toplevel.append(folder)
        return folder

    def register_piece(self, name, piece):
        """
        Registers a :class:`~puzzlepiece.piece.Piece` object with the Puzzle.
        This is done by default when a :class:`~puzzlepiece.piece.Piece` is added with
        :func:`~puzzlepiece.puzzle.Puzzle.add_piece`, :func:`puzzlepiece.puzzle.Folder.add_piece`,
        or :func:`puzzlepiece.puzzle.Grid.add_piece`, so this method should rarely be called manually.
        """
        self.pieces[name] = piece
        piece.setTitle(name)

    # Other methods

    def process_events(self):
        """
        Forces the QApplication to process events that happened while a callback was executing.
        Can for example update plots while a long process is running, or run any keyboard
        shortcuts pressed while proecessing.
        """
        self.app.processEvents()

    _shutdown_threads = QtCore.Signal()

    def run_worker(self, worker):
        """
        Add a Worker to the Puzzle's Threadpool and runs it. See :class:`puzzlepiece.threads`
        for more details on how to set up a Worker.
        """
        if hasattr(worker, "stop"):
            # This signal is emitted when the application is shutting down,
            # so we're telling the LiveWorker to stop
            self._shutdown_threads.connect(worker.stop)
        self._threadpool.start(worker)

    def _excepthook(self, exctype, value, traceback):
        self._old_excepthook(exctype, value, traceback)

        # Stop any threads that may be running
        self._shutdown_threads.emit()

        # Only do custom exception handling in the main thread, otherwise the messagebox
        # or other such things are likely to break things.
        if QtCore.QThread.currentThread() == self.app.thread():
            self.custom_excepthook(exctype, value, traceback)

            box = QtWidgets.QMessageBox()
            box.setText(str(value) + "\n\nCheck console for details.")
            box.exec()

    def custom_excepthook(self, exctype, value, traceback):
        """
        Override or replace this method to call a custom handler whenever an exception is raised.
        This will run after the defatult exception handler (``sys.__excepthook__``), but before a
        GUI alert is displayed.
        """
        pass

    # Convenience methods

    def _docs(self):
        dialog = QtWidgets.QDialog(self)
        layout = QtWidgets.QVBoxLayout()
        tree = QtWidgets.QTreeWidget()
        tree.setHeaderLabels(("pieces", "get?", "set?"))

        def copy_item(item):
            if hasattr(item, "puzzlepiece_descriptor"):
                self.app.clipboard().setText(item.puzzlepiece_descriptor)

        tree.itemDoubleClicked.connect(copy_item)

        for piece_name in self.pieces:
            piece_item = QtWidgets.QTreeWidgetItem(tree, (piece_name,))

            # First, params
            tree_item = QtWidgets.QTreeWidgetItem(piece_item, ("params",))
            for param_name in self.pieces[piece_name].params:
                param = self.pieces[piece_name].params[param_name]
                G = "⟳" if param._getter is not None else ""
                S = "✓" if param._setter is not None else ""
                param_item = QtWidgets.QTreeWidgetItem(tree_item, (param_name, G, S))
                param_item.puzzlepiece_descriptor = "{}:{}".format(
                    piece_name, param_name
                )

            # Then, actions
            tree_item = QtWidgets.QTreeWidgetItem(piece_item, ("actions",))
            for action_name in self.pieces[piece_name].actions:
                action = self.pieces[piece_name].actions[action_name]
                action_item = QtWidgets.QTreeWidgetItem(tree_item, (action_name,))
                action_item.puzzlepiece_descriptor = "{}:{}".format(
                    piece_name, action_name
                )

                button = QtWidgets.QToolButton()
                icon = self.style().standardIcon(
                    QtWidgets.QStyle.StandardPixmap.SP_MediaPlay
                )
                button.setIcon(icon)
                button.clicked.connect(lambda x=False, action=action: action())
                tree.setItemWidget(action_item, 1, button)

        for i in range(0, 3):
            tree.header().setSectionResizeMode(
                i, QtWidgets.QHeaderView.ResizeMode.ResizeToContents
            )

        label = QtWidgets.QLabel()
        label.setText(
            "Double-click on any row to copy the param/action identifier for use in scripts."
        )
        label.setWordWrap(True)

        layout.addWidget(tree)
        layout.addWidget(label)

        dialog.setLayout(layout)
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()

    def _export_setup(self):
        dialog = QtWidgets.QDialog(self)
        layout = QtWidgets.QVBoxLayout()

        label = QtWidgets.QLabel()
        label.setText(
            "The script below sets all currently params that don't have setters or getters to the current values."
        )
        label.setWordWrap(True)
        layout.addWidget(label)

        text = ""
        for piece_name in self.pieces:
            keys = self.pieces[piece_name].params
            for key in keys:
                param = self.pieces[piece_name].params[key]
                if param.visible and param._setter is None and param._getter is None:
                    text += "set:{}:{}:{}\n".format(piece_name, key, param.get_value())

        text_box = QtWidgets.QPlainTextEdit()
        text_box.setPlainText(text)
        layout.addWidget(text_box)

        button = QtWidgets.QPushButton("Save")

        def __save_export():
            fname = str(QtWidgets.QFileDialog.getSaveFileName(self, "Save file...")[0])
            with open(fname, "w") as f:
                f.write(text_box.toPlainText())

        button.clicked.connect(__save_export)
        layout.addWidget(button)

        dialog.setLayout(layout)
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()

    def _call_stop(self):
        for piece_name in self.pieces:
            self.pieces[piece_name].call_stop()
        self._shutdown_threads.emit()

    def _button_layout(self):
        layout = QtWidgets.QHBoxLayout()

        for function, icon, text in zip(
            (self._docs, self._export_setup, self._call_stop),
            (
                QtWidgets.QStyle.StandardPixmap.SP_MessageBoxInformation,
                QtWidgets.QStyle.StandardPixmap.SP_DialogSaveButton,
                QtWidgets.QStyle.StandardPixmap.SP_BrowserStop,
            ),
            ("Tree (F1)", "Export (F2)", "STOP (F3)"),
        ):
            button = QtWidgets.QPushButton(text)
            icon = self.style().standardIcon(icon)
            button.setIcon(icon)
            button.clicked.connect(lambda x=False, action=function: action())
            layout.addWidget(button)

        return layout

    def __getitem__(self, name):
        return self.pieces[name]

    def _ipython_key_completions_(self):
        values = list(self.pieces.keys())
        for piece in self.pieces.keys():
            values.extend([f"{piece}:{param}" for param in self.pieces[piece].params])
        return values

    def run(self, text):
        """
        Execute script commands for this Puzzle as described in :func:`puzzlepiece.parse.run`.
        """
        parse.run(text, self)

    def get_values(self, text):
        """
        Get the values from multiple params as a list.

        :param text: A string of comma-separated param strings
          as described in :func:`puzzlepiece.parse.parse_params`.
        :rtype: list
        """
        return [param.get_value() for param in parse.parse_params(text, self)]

    def record_values(self, text, dictionary=None):
        """
        Get the values from multiple params and record them in a dictionary.
        Useful for storing metadata about a measurement.

        :param text: A string of comma-separated param strings
          as described in :func:`puzzlepiece.parse.parse_params`.
        :param dictionary: If provided, this function will write the param names
          and values to this dictionary. Otherwise, a new one is created and returned.
        :rtype: dict
        """
        params = parse.parse_params(text, self)
        names = text.split(", ")

        if dictionary is None:
            dictionary = {}

        for name, param in zip(names, params):
            dictionary[name] = param.get_value()

        return dictionary

    # Qt overrides

    def keyPressEvent(self, event):
        """
        Pass down keypress events to child Pieces and Folders.
        Overwrites a QT method.

        :meta private:
        """
        if event.key() == QtCore.Qt.Key.Key_F1:
            self._docs()
        elif event.key() == QtCore.Qt.Key.Key_F2:
            self._export_setup()
        elif event.key() == QtCore.Qt.Key.Key_F3:
            self._call_stop()
        for widget in self._toplevel:
            widget.handle_shortcut(event)

    _close_popups = QtCore.Signal()

    def closeEvent(self, event):
        """
        Tell the Pieces the window is closing, so they can for example disconnect hardware.
        Overwrites a QT method.

        :meta private:
        """
        self._shutdown_threads.emit()
        self._close_popups.emit()

        if not self.debug:
            for piece_name in self.pieces:
                self.pieces[piece_name].handle_close(event)

        # Reinstate the original excepthook
        sys.excepthook = self._old_excepthook
        super().closeEvent(event)


QApp = QtWidgets.QApplication
"""A QApplication has to be constructed before any Qt objects
(including the Puzzle and the Pieces), so this is a convenient shortcut to
the QApplication class (see https://doc.qt.io/qt-6/qapplication.html).
"""


class Folder(QtWidgets.QTabWidget):
    """
    A tabbed group of :class:`~puzzlepiece.puzzle.Piece` or :class:`~puzzlepiece.puzzle.Grid`
    objects within the :class:`~puzzlepiece.puzzle.Puzzle`.

    Best created with :func:`puzzlepiece.puzzle.Puzzle.add_folder`.
    """

    def __init__(self, puzzle, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.puzzle = puzzle
        self.pieces = []

    def add_piece(self, name, piece):
        """
        Adds a :class:`~puzzlepiece.piece.Piece` as a tab to this Folder, and registers it with the
        parent :class:`~puzzlepiece.puzzle.Puzzle`.

        :param name: Identifying string for the Piece.
        :param piece: A :class:`~puzzlepiece.piece.Piece` object or a class defining one (which will
          be automatically instantiated).
        :rtype: puzzlepiece.piece.Piece
        """
        if isinstance(piece, type):
            piece = piece(self.puzzle)
        self.addTab(piece, name)
        self.pieces.append(piece)
        self.puzzle.register_piece(name, piece)
        piece.folder = self

        # No title or border displayed when Piece in Folder
        piece.setTitle(None)
        piece.setStyleSheet("QGroupBox {border:0;}")
        # Remove most of the border if the stylesheet fails
        piece.setFlat(True)

        return piece

    def add_grid(self, name):
        """
        Adds a :class:`~puzzlepiece.puzzle.Grid` as a tab to this Folder.

        :param name: Identifying string for the :class:`~puzzlepiece.puzzle.Grid`.
        :rtype: puzzlepiece.puzzle.Grid
        """

        grid = Grid(self.puzzle)

        self.addTab(grid, name)
        self.pieces.append(grid)
        grid.folder = self

        return grid

    def handle_shortcut(self, event):
        """
        Pass down keypress events only to the **active** child :class:`~puzzlepiece.piece.Piece` or
        :class:`~puzzlepiece.puzzle.Grid`.

        :meta private:
        """
        self.currentWidget().handle_shortcut(event)

    def _replace_piece(self, name, old_piece, new_piece):
        if old_piece in self.pieces:
            index = self.indexOf(old_piece)
            self.insertTab(index, new_piece, name)
            new_piece.folder = self
            # No title or border displayed when Piece in Folder
            new_piece.setTitle(None)
            new_piece.setStyleSheet("QGroupBox {border:0;}")
            new_piece.setFlat(True)

            self.pieces.remove(old_piece)
            self.pieces.append(new_piece)
        else:
            for widget in self.pieces:
                if isinstance(widget, Grid):
                    widget._replace_piece(name, old_piece, new_piece)


class Grid(QtWidgets.QWidget):
    """
    A grid layout for :class:`~puzzlepiece.piece.Piece` objects. For when you need multiple Pieces
    within a single :class:`~puzzlepiece.puzzle.Folder` tab.

    Best created with :func:`puzzlepiece.puzzle.Puzzle.add_folder`.
    """

    def __init__(self, puzzle, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.puzzle = puzzle
        self.pieces = []
        self.layout = QtWidgets.QGridLayout()
        self.setLayout(self.layout)

    def add_piece(self, name, piece, row, column, rowspan=1, colspan=1):
        """
        Adds a :class:`~puzzlepiece.piece.Piece` to the grid layout, and registers it with the parent
        :class:`~puzzlepiece.puzzle.Puzzle`.

        :param name: Identifying string for the Piece.
        :param piece: A :class:`~puzzlepiece.piece.Piece` object or a class defining one (which will
          be automatically instantiated).
        :param row: Row index for the grid layout.
        :param column: Column index for the grid layout.
        :param rowspan: Height in rows.
        :param column: Width in columns.
        :rtype: puzzlepiece.piece.Piece
        """
        if isinstance(piece, type):
            piece = piece(self.puzzle)
        self.layout.addWidget(piece, row, column, rowspan, colspan)
        self.pieces.append(piece)
        self.puzzle.register_piece(name, piece)
        piece.folder = self

        return piece

    def _replace_piece(self, name, old_piece, new_piece):
        if old_piece in self.pieces:
            self.layout.replaceWidget(old_piece, new_piece)
            new_piece.setTitle(name)
            new_piece.folder = self
            self.pieces.remove(old_piece)
            self.pieces.append(new_piece)

    def handle_shortcut(self, event):
        """
        Pass down keypress events to child Pieces.

        :meta private:
        """
        for widget in self.pieces:
            widget.handle_shortcut(event)

    def setCurrentWidget(self, piece):
        """
        Passes the `setCurrentWidget` to this Grid's Folder, allowing Pieces within this Grid
        to correctly call :func:`puzzlepiece.piece.Piece.elevate`.

        :meta private:
        """
        if self.folder is not None:
            self.folder.setCurrentWidget(self)


class PieceDict:
    """
    A dictionary wrapper that enforces single-use of keys, and raises a more useful error when
    a Piece tries to use another Piece that hasn't been registered.

    It also allows indexing params directly by using this key format: ``[piece_name]:[param_name]``.
    """

    def __init__(self):
        self._dict = {}

    def __setitem__(self, key, value):
        if key in self._dict:
            raise KeyError("A Piece with id '{}' already exists".format(key))
        self._dict[key] = value

    def __iter__(self):
        for key in self._dict:
            yield key

    def __getitem__(self, key):
        if key not in self._dict:
            try:
                piece, param = key.split(":")
                return self._dict[piece][param]
            except ValueError:
                # key is not in the piece:param format
                pass
            raise KeyError(
                "A Piece with id '{}' is required, but doesn't exist".format(key)
            )
        return self._dict[key]

    def _replace_item(self, key, value):
        self._dict[key] = value

    def __contains__(self, item):
        return item in self._dict

    def keys(self):
        return self._dict.keys()

    def __repr__(self):
        return "PieceDict({})".format(", ".join(self._dict.keys()))


class Globals:
    """
    A dictionary wrapper used for :attr:`puzzlepiece.puzzle.Puzzle.globals`. It behaves like
    a dictionary, allowing :class:`puzzlepiece.piece.Piece` objects to share device APIs
    with each other.

    Additionally, :func:`~puzzlepiece.puzzle.Globals.require` and
    :func:`~puzzlepiece.puzzle.Globals.release` can be used to keep track of the Pieces
    using a given variable, so that the API can be loaded once and then unloaded once
    all the Pieces are done with it.
    """

    def __init__(self):
        self._dict = {}
        self._counts = {}

    def require(self, name):
        """
        Register that a Piece is using the variable with a given name. This will increase
        an internal counter to indicate the Piece having a hold on the variable.

        Returns `False` if this is the first time a variable is being registered (and thus
        setup is needed) or `True` if the variable has been registered already.

        For example, this can be used within :func:`~puzzlepiece.piece.Piece.setup`::

            def setup(self):
                if not self.puzzle.globals.require('sdk'):
                    # Load the SDK if not done already by a different Piece
                    self.puzzle.globals['sdk'] = self.load_sdk()

        :param name: a dictionary key for the required variable
        :rtype: bool
        """
        if name not in self._dict:
            self._dict[name] = None
            self._counts[name] = 1
            return False
        else:
            self._counts[name] += 1
            return True

    def release(self, name):
        """
        Indicate that a Piece is done using the variable with a given name.
        This will decrease an internal counter to indicate the Piece is releasing
        its hold on the variable.

        Returns `False` if the counter is non-zero (so different Pieces are still using
        this variable) or `True` if all Pieces are done with the variable (in that case
        the SDK can be safely shut down for example).

        For example, this can be used within :func:`~puzzlepiece.piece.Piece.handle_close`::

            def handle_close(self):
                if self.puzzle.globals.release('sdk'):
                    # Unload the SDK if all Pieces are done with it
                    self.puzzle.globals['sdk'].stop()

        :param name: a dictionary key for the variable being released
        :rtype: bool
        """
        if name not in self._dict:
            raise KeyError(f"No global variable with id '{name}' to release")
        if name not in self._counts:
            raise KeyError(
                f"Cannot release '{name}' since it hasn't been registered with 'require'"
            )
        self._counts[name] -= 1
        return self._counts[name] < 1

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __getitem__(self, key):
        if key not in self._dict:
            raise KeyError("No global variable with id '{}'".format(key))
        return self._dict[key]

    def __delitem__(self, key):
        del self._dict[key]
        if key in self._counts:
            del self._counts[key]

    def __contains__(self, item):
        return item in self._dict

    def keys(self):
        return self._dict.keys()

    def __repr__(self):
        return "Globals({})".format(", ".join(self._dict.keys()))


class PretendPuzzle:
    """
    A placeholder object used if no :class:`~puzzlepiece.puzzle.Puzzle` is provided
    when creating a :class:`puzzlepiece.puzzle.Piece`. Its `debug` attribute is
    always True.
    """

    debug = True

    def process_events(self):
        """
        Like :func:`puzzlepiece.puzzle.Puzzle.process_events()`.
        """
        QtWidgets.QApplication.instance().processEvents()



from qtpy import QtCore, QtWidgets
from functools import partial
import time


class CallLater:
    """
    A **callable** object that will run a specified function when the next Qt event loop iteration
    occurs, but only once, irrespective of how many times the ``CallLater`` was called since the
    last iteration of the event loop.

    This is useful if you are stuck in a long running function and want to schedule something for
    when it finishes or when :func:`puzzlepiece.puzzle.Puzzle.process_events` is called.

    Let's say a :class:`puzzlepiece.piece.Piece` has an image view defined in
    :func:`~puzzlepiece.piece.Piece.custom_layout` and we'd like to update the image when the
    Piece's ``image`` :class:`~puzzlepiece.param.ParamArray` changes. We can then do::

        def update_image():
            image.setImage(self.params['image'].value)
        update_later = pzp.threads.CallLater(update_image)
        self.params['image'].changed.connect(update_later)

    instead of directly connecting ``update_image``. This way the image data will be set only once, when
    the next GUI update occurs. So the code is more efficient if somewhere else we do::

        for i in range(1000):
            self.params['image'].get_value()
            # The image view won't update here as the Qt event loop is stuck inside this function
        puzzle.process_events()
        # update_image will be called once as part of this event loop iteration,
        # instead of a 1000 times.
        # You can also directly call the CallLater object:
        update_later()

    This behaviour is implemented through a single shot QTimer stored internally.

    :param function: The function to be called.
    :param args and kwargs: further arguments and keyword arguments can be provided,
      they will be passed to the function when it is called.
    """

    def __init__(self, function, *args, **kwargs):
        self._timer = QtCore.QTimer()
        self._timer.setSingleShot(True)
        self._timer.setTimerType(QtCore.Qt.TimerType.PreciseTimer)
        self._timer.timeout.connect(partial(function, *args, **kwargs))

    def __call__(self, *args, **kwargs):
        self._timer.start()


class _Emitter(QtCore.QObject):
    # The Emitter is needed as a QRunnable is not a QObject, and cannot emit it's own signals.
    # So we set up the Signal here, and let a Worker instance an Emitter for its use
    signal = QtCore.Signal(object)


class Worker(QtCore.QRunnable):
    """
    Generic worker (QRunnable) that calls a function in a thread.

    :param function: The function for the Worker to run.
    :param args: list of arguments to be forwarded to the function when run.
    :param kwargs: dictionary of keyword arguments to be forwarded to the function when run.
    """

    def __init__(self, function, args=None, kwargs=None):
        self.function = function
        self.args = args if args is not None else []
        self.kwargs = kwargs if kwargs is not None else {}
        #: Bool flag, True when task finished.
        self.done = False
        self._emitter = _Emitter()
        #: A Qt signal emitted when the function returns, passes the returned value to the connected Slot.
        self.returned = self._emitter.signal
        super().__init__()

    @QtCore.Slot()
    def run(self):
        """
        Run the Worker. Shouldn't be excecuted directly, instead
        use :func:`puzzlepiece.puzzle.Puzzle.run_worker`.
        """
        try:
            r = self.function(*self.args, **self.kwargs)
            self.returned.emit(r)
        finally:
            self.done = True


class _Done_Emitter(QtCore.QObject):
    signal = QtCore.Signal()


class LiveWorker(Worker):
    """
    A Worker that calls a function repeatedly in a thread,
    separated by specified time intervals.

    :param function: The function for the Worker to run.
    :param sleep: Time to sleep between function calls in seconds.
    :param args: list of arguments to be forwarded to the function when run.
    :param kwargs: dictionary of keyword arguments to be forwarded to the function when run.
    """

    def __init__(self, function, sleep=0.1, args=None, kwargs=None):
        self.stopping = False
        self.sleep = sleep
        super().__init__(function, args, kwargs)
        #: A Qt signal emitted each time the function returns, passes the returned value to the connected Slot.
        self.returned = self.returned
        # The above line is there for documentation to compile correctly
        self._done_emitter = _Done_Emitter()
        #: A Qt signal emitted when the LiveWorker is stopped.
        self.done_signal = self._done_emitter.signal

    def stop(self):
        """
        Ask the Worker to stop. This will only take effect once the current
        execution of the function is over.
        """
        self.stopping = True

    @QtCore.Slot()
    def run(self):
        """
        Run the Worker. Shouldn't be excecuted directly, instead
        use :func:`puzzlepiece.puzzle.Puzzle.run_worker`.
        """
        try:
            while not self.stopping:
                r = self.function(*self.args, **self.kwargs)
                self.returned.emit(r)
                if not self.stopping:
                    time.sleep(self.sleep)
        finally:
            self.done = True
            self.done_signal.emit()


class PuzzleTimer(QtWidgets.QWidget):
    """
    A Widget that displays a checkbox. When the checkbox is checked, an associated
    function is called repeatedly in a thread, separated by specified time intervals.

    :param name: The display name for this PuzzleTimer.
    :param puzzle: The parent :class:`~puzzlepiece.puzzle.Puzzle` (needed as the Puzzle
      is charged with the QThreadPool that runs the tasks).
    :param function: The function for the PuzzleTimer to run.
    :param sleep: Time to sleep between function calls in seconds.
    :param args: list of arguments to be forwarded to the function when run.
    :param kwargs: dictionary of keyword arguments to be forwarded to the function when run.
    """

    #: A Qt signal emitted each time the associated LiveWorker returns, passes the returned value to the connected Slot.
    returned = QtCore.Signal(object)

    def __init__(self, name, puzzle, function, sleep=0.1, args=None, kwargs=None):
        self.name = name
        self.puzzle = puzzle
        self.function = function
        self._sleep = sleep
        self.args = args
        self.kwargs = kwargs
        self.worker = None

        super().__init__()

        layout = QtWidgets.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self.input = QtWidgets.QCheckBox(self.name)
        layout.addWidget(self.input)
        self.input.stateChanged.connect(self._state_handler)

    def _state_handler(self, state):
        if state and (self.worker is None or self.worker.done):
            self.worker = LiveWorker(self.function, self._sleep, self.args, self.kwargs)
            self.worker.returned.connect(self._return_handler)
            self.worker.done_signal.connect(self.stop)
            self.puzzle.run_worker(self.worker)
        elif self.worker is not None:
            self.worker.stop()

    def _return_handler(self, value):
        self.returned.emit(value)

    def stop(self):
        """
        Ask the PuzzleTimer to stop.

        This will only take effect once the current execution of the function is over,
        see :func:`puzzlepiece.threads.LiveWorker.stop`.
        """
        self.input.setChecked(False)

    @property
    def sleep(self):
        """
        Interval in seconds between executions of the associated function.
        The property can be modified to change the interval for the associated Worker,
        even if it's already running.
        """
        return self._sleep

    @sleep.setter
    def sleep(self, value):
        if self.worker is not None:
            self.worker.sleep = float(value)
        self._sleep = float(value)



from qtpy import QtWidgets, QtCore

import puzzlepiece as pzp


class DataGrid(QtWidgets.QWidget):
    """
    A table containing multiple :class:`~puzzlepiece.extras.datagrid.Row` objects, each acting kind of like
    a :class:`~puzzlepiece.piece.Piece` object, in that it has params and actions
    (see :class:`~puzzlepiece.param.BaseParam` and :class:`~puzzlepiece.action.Action`).

    This is a QWidget, co it can be added to your Piece's :func:`~puzzlepiece.piece.Piece.custom_layout`
    or used as a standalone Widget if you know what you're doing.

    **This is not very performant, very large numbers of Rows should be avoided!** Consider using
    Qt's QTableWidget instead.

    :param row_class: The :class:`~puzzlepiece.extras.datagrid.Row` class that will be used to construct Rows.
    :param puzzle: (optional) The parent :class:`~puzzlepiece.puzzle.Puzzle`.
    """

    #: A Qt signal emitted when a row is added or removed.
    rows_changed = QtCore.Signal()
    #: A Qt signal emitted when any data in the DataGrid changes (including when rows are added/removed).
    data_changed = QtCore.Signal()

    def __init__(self, row_class, puzzle=None, parent_piece=None):
        super().__init__()
        #: Reference to the parent :class:`~puzzlepiece.puzzle.Puzzle`.
        self.puzzle = puzzle or pzp.puzzle.PretendPuzzle()
        self.parent_piece = parent_piece
        self._row_class = row_class
        row_example = row_class(self.puzzle)
        self.param_names = row_example.params.keys()

        self._tree = QtWidgets.QTreeWidget()
        if len(row_example.actions.keys()):
            self._tree.setHeaderLabels(("ID", *row_example.params.keys(), "actions"))
        else:
            self._tree.setHeaderLabels(("ID", *row_example.params.keys()))

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self._tree)
        self.setLayout(layout)

        #: A list of Rows.
        self.rows = []
        self._items = []
        self._root = self._tree.invisibleRootItem()
        self._slots = {}
        self.rows_changed.connect(self.data_changed.emit)

    @property
    def values(self):
        """
        The current values for all params in the Rows (this does not invoke their getters).
        """
        return [{key: x[key].value for key in x.params} for x in self.rows]

    def add_row(self, row_class=None, **kwargs):
        """
        Add a Row with default param values.

        :param row_class: row class to use, if not specified uses the one provided
          when the DataGrid was created. The class provided here should have the same
          params and actions as the original one!
        :param kwargs: keyword arguments matching param names can be passed
          to set param values in the new row
        """
        item = QtWidgets.QTreeWidgetItem(self._tree, (str(len(self.rows)),))
        row_class = row_class or self._row_class
        row = row_class(self, self.puzzle)
        row._populate_item(self._tree, item)
        self.rows.append(row)
        self._items.append(item)
        for key in kwargs:
            row.params[key].set_value(kwargs[key])
        for param_name in row.params:
            if param_name in self._slots:
                for slot in self._slots[param_name]:
                    row.params[param_name].changed.connect(slot)
            row.params[param_name].changed.connect(self.data_changed.emit)
        self.rows_changed.emit()
        return row

    def remove_row(self, id):
        """
        Remove the row with the given id.

        :param id: id of the row to remove.
        """
        self._root.removeChild(self._items[id])
        del self._items[id]
        del self.rows[id]
        for i in range(len(self.rows)):
            self._items[i].setText(0, str(i))
        self.rows_changed.emit()

    def select_row(self, id):
        self._tree.setCurrentItem(self._items[id])

    def get_index(self, row):
        """
        Get the current index of a given :class:`~puzzlepiece.extras.datagrid.Row` object.

        :param row: the row object.
        :rtype: int
        """
        return self.rows.index(row)

    def clear(self):
        """
        Remove all rows.
        """
        self._tree.clear()
        self.rows = []
        self._items = []
        self.rows_changed.emit()

    def add_changed_slot(self, param_name, function):
        """
        Connect a Slot (usually a method) to the :attr:`~puzzlepiece.param.BaseParam.changed`
        Signal of the given param in all the rows (including Rows added in the future).

        :param param_name: The name of the param whose changed Signal we're connecting to.
        :param function: any method or other Qt Slot to connect.
        """
        if param_name in self._slots:
            self._slots[param_name].append(function)
        else:
            self._slots[param_name] = [function]
        for row in self.rows:
            row.params[param_name].changed.connect(function)


class Row:
    """
    A Row is a template for a :class:`~puzzlepiece.extras.datagrid.DataGrid` object, holding the data (params)
    and actions that the DataGrid displays.

    It acts kind of like a :class:`~puzzlepiece.piece.Piece` object, in that it has params and actions
    (see :class:`~puzzlepiece.param.BaseParam` and :class:`~puzzlepiece.action.Action`).

    :param parent: (optional) The parent DataGrid.
    :parem puzzle: (optional) The parent Puzzle.
    """

    def __init__(self, parent=None, puzzle=None):
        self.puzzle = puzzle or pzp.puzzle.PretendPuzzle()
        self.parent = parent
        #: dict: A dictionary of this Row's params (see :class:`~puzzlepiece.param.BaseParam`). You can also directly index the Row object with the param name.
        self.params = {}
        #: dict: A dictionary of this Row's actions (see :class:`~puzzlepiece.action.Action`).
        self.actions = {}
        self.define_params()
        self.define_actions()
        for key in self.params:
            self.params[key]._main_layout.removeWidget(self.params[key].label)

    def define_params(self):
        """
        Override to define params using decorators from :mod:`puzzlepiece.param`.
        """
        pass

    def define_actions(self):
        """
        Override to define actions using decorators from :mod:`puzzlepiece.action`.
        """
        pass

    def open_popup(self, popup, name=None):
        """
        Open a popup window for this Row.
        See :func:`puzzlepiece.piece.Piece.open_popup`.

        :param popup: a :class:`puzzlepiece.piece.Popup` _class_ to instantiate
        :param name: text to show as the window title
        :rtype: puzzlepiece.piece.Popup
        """
        # Instantiate the popup
        if isinstance(popup, type):
            popup = popup(self, self.puzzle)
        popup.setStyleSheet("QGroupBox {border:0;}")

        # Make a dialog window for the popup to live in
        dialog = pzp.piece._QDialog(self.parent, popup)
        layout = QtWidgets.QVBoxLayout()
        dialog.setLayout(layout)
        layout.addWidget(popup)
        dialog.setWindowTitle(name or "Popup")

        # Display the dialog
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()
        self.puzzle._close_popups.connect(dialog.accept)

        return popup

    def elevate(self):
        """
        For compatibility with the Piece's API.

        :meta private:
        """
        pass

    def _populate_item(self, tree, item):
        visible_params = [x for x in self.params if self.params[x].visible]
        for i, key in enumerate(visible_params):
            tree.setItemWidget(item, i + 1, self.params[key])
        if len(self.actions.keys()):
            tree.setItemWidget(item, i + 2, self._action_buttons())

    def _action_buttons(self):
        widget = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout()
        widget.setLayout(layout)

        visible_actions = [key for key in self.actions if self.actions[key].visible]
        for i, key in enumerate(visible_actions):
            button = QtWidgets.QPushButton(key)
            button.clicked.connect(lambda x=False, _key=key: self.actions[_key]())
            layout.addWidget(button)
        return widget

    def __iter__(self):
        for key in self.params:
            yield key

    def __getitem__(self, key):
        return self.params[key]

    def __contains__(self, item):
        return item in self.params



from qtpy import QtWidgets, QtCore, QtGui
from functools import wraps
import numpy as np


_red_bg_palette = QtGui.QPalette()
_red_bg_palette.setColor(
    _red_bg_palette.ColorRole.Window, QtGui.QColor(252, 217, 202, 255)
)


class BaseParam(QtWidgets.QWidget):
    """
    A param object represents a gettable/settable value within a :class:`~puzzlepiece.piece.Piece`. This can be
    a general variable for some operation, a setting of a physical device (like laser power), a readout from a
    physical device (counts from a spectrometer), or something that can be measured and set (the position of a
    movable stage).

    A general param is just a variable with a GUI presence (if needed) that can be read from and written to by
    Pieces. It can then have getter and setter functions that are called when you get/set the value of the param.

    To set and get the value of the param, use the :func:`~puzzlepiece.param.BaseParam.set_value` and
    :func:`~puzzlepiece.param.get_value` methods.

    **In most cases, you will not instantiate a param directly, but rather use decorators defined within this module
    (like** :func:`puzzlepiece.param.spinbox` **)
    to register params within a Piece's** :func:`~puzzlepiece.piece.Piece.define_params` **method.**

    :param name: A unique (per Piece) id name for the param
    :param value: Default value (can be None)
    :param setter: A function to be called when setting the param, taking the value as argument. It may return a new value.
    :param getter: A function to be called when obtaining the value of the param, raturning the new value.
    :param visible: If True, the Piece will generate display a GUI component for the param.
    :param format: Default format for displaying the param if a custom input is not defined, and in
      :func:`puzzplepiece.parse.format`. For example `{:.2f}`, see https://pyformat.info/ for further details.
    :param _type: int, float, str etc - the type of the param value. Can be inferred if a default value is passed.
    """

    #: A Qt signal emitted when the value changes
    changed = QtCore.Signal()
    _type = None

    def __init__(
        self,
        name,
        value,
        setter=None,
        getter=None,
        visible=True,
        format="{}",
        _type=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._name = name
        self._setter = setter
        self._getter = getter
        self._value = None
        self._visible = visible
        self._format = format
        if _type is not None:
            self._type = _type
        if self._type is None:
            if value is not None:
                # Infer type if not provided by subclassing
                self._type = type(value)
            else:
                # Bypass type conversion if type not known
                self._type = lambda x: x

        self._main_layout = layout = QtWidgets.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self.setPalette(_red_bg_palette)

        # Give the param a label
        self.label = QtWidgets.QLabel()
        self.label.setText(name + ":")
        layout.addWidget(self.label, 0, 0)

        # Give the param an input box
        self.input, make_set_button = self._make_input(
            value, self._value_change_handler
        )
        if self._setter is None and value is not None:
            self._value = self._type(value)
        if self._value is None:
            # Highlight that the setter or getter haven't been called yet
            self.setAutoFillBackground(True)
        layout.addWidget(self.input, 0, 1)
        # self.set_value(value)

        # Give it buttons for setting and getting the value
        if self._getter is not None:
            self._make_get_button()
        if self._setter is not None and make_set_button:
            self._make_set_button()

    def _make_set_button(self):
        self._set_button = QtWidgets.QToolButton()
        # Icon list: https://www.pythonguis.com/faq/built-in-qicons-pyqt/
        icon = self.style().standardIcon(
            QtWidgets.QStyle.StandardPixmap.SP_DialogApplyButton
        )
        self._set_button.setIcon(icon)
        # Using a lambda as the clicked signal always passes False as the first argument
        self._set_button.clicked.connect(lambda x: self.set_value())
        self._main_layout.addWidget(self._set_button, 0, 3)

    def _make_get_button(self):
        self._get_button = QtWidgets.QToolButton()
        icon = self.style().standardIcon(
            QtWidgets.QStyle.StandardPixmap.SP_BrowserReload
        )
        self._get_button.setIcon(icon)
        self._get_button.clicked.connect(lambda x: self.get_value())
        self._main_layout.addWidget(self._get_button, 0, 2)

    def _value_change_handler(self):
        if self._setter is not None:
            # Highlight the param box if a setter is set
            self.setAutoFillBackground(True)
        else:
            # If there's no setter, we call set_value to set the value from input
            self.set_value()

    def set_value(self, value=None):
        """
        Set the value of the param. If a setter is registered, it will be called.

        If the setter returns a value, this will become the new value of this param.

        If the setter doesn't return a value, the getter will be called if present,
        and the value it returns will become the new value of this param.

        If the setter doesn't return a value, and there is no getter, the value
        provided as an argument will become the new value of this param

        :param value: The value this param should be set to (if None, we grab the value from
          the param's input box.)
        :returns: The new value of the param.
        """
        # If a value is not provided, grab one from the input
        if value is None:
            value = self._input_get_value()
        else:
            # Otherwise push the given value to the input
            value = self._type(value)
            self._input_set_value(value)

        if self._setter is not None:
            # Call setter if it exists. It may return a new value.
            new_value = self._setter(value)
            if new_value is None:
                # If the setter did not return a value, see if there is a getter
                if self._getter is not None:
                    new_value = self._getter()
                    new_value = self._type(new_value)
                else:
                    # Otherwise the new value is just the value we're setting
                    new_value = value
            # Update the value stored to the new value
            self._value = new_value
            # Update the input as well
            self._input_set_value(new_value)
        else:
            self._value = value

        # Clear the highlight and emit the changed signal
        self.setAutoFillBackground(False)
        self.changed.emit()
        return self._value

    def get_value(self):
        """
        Get the value for this param. If a getter is registered, it will be called, and the
        returned value becomes the params new value, and is returned by this method.

        :returns: Value of the param (retuned by the getter if registered, otherwise the value
          currently stored by the param).
        """
        if self._getter is not None:
            new_value = self._getter()
            new_value = self._type(new_value)
            self._value = new_value

            # Set the value to the input and emit signal if needed
            self._input_set_value(new_value)
            self.setAutoFillBackground(False)
            self.changed.emit()

            return new_value
        else:
            return self._value

    def set_setter(self, piece):
        """
        Create a decorator to register a setter for this param. This would be used within
        :func:`puzzlepiece.piece.Piece.define_params` as ``@param.set_setter(self)`` decorating
        a function.
        """

        def decorator(setter):
            wrapper = wrap_setter(piece, setter)
            self._setter = wrapper
            self._make_set_button()
            return self

        return decorator

    def set_getter(self, piece):
        """
        Create a decorator to register a getter for this param. This would be used within
        :func:`puzzlepiece.piece.Piece.define_params` as ``@param.set_getter(self)`` decorating
        a function.
        """

        def decorator(getter):
            wrapper = wrap_getter(piece, getter)
            self._getter = wrapper
            self._make_get_button()
            return self

        return decorator

    @property
    def value(self):
        """
        Current, internally stored value.
        The getter is not called when this property is accessed.
        """
        return self._value

    def _make_input(self, value=None, connect=None):
        """
        Create an input box for the GUI display of this param. This should be overriden to implement
        custom param display types (like the spinboxes, text inputs, and checkboxes provided by default).

        :param value: The value to put into the input box by default.
        :param connect: A function that will be connected to the value changed signal of the input.
        :rtype: (QWidget - input box, bool - whether a setter button is needed)

        :meta public:
        """
        input = QtWidgets.QLabel()
        if value is not None:
            input.setText(self._format.format(value))
        return input, True

    def _input_set_value(self, value):
        """
        Set the value of the input box. This should be overriden to implement
        custom param display types.

        As this is a low-level method used internally, it **should not** emit valueChanged
        signals for its input box. To stop this from happening, use Qt's `blockSignals`
        method::

            self.input.blockSignals(True)
            self.input.setText(value)
            self.input.blockSignals(False)

        :meta public:
        """
        self.input.setText(self._format.format(value))

    def _input_get_value(self):
        """
        Set the value of the input box. This should be overriden to implement
        custom param display types.

        :rtype: Should return the velue with correct type, as specified by `self._type`

        :meta public:
        """
        return self._type(self.input.text())

    def make_child_param(self, kwargs=None):
        """
        Create and return a child param. Changing the value of the child changes the value
        of the parent, but not vice versa - each child has a getter that allows for refreshing
        the value from the parent.

        The child will be of the same type as the parent - a checkbox, spinbox, etc.

        The parent's getter will be called when you :func:`~puzzlepiece.param.BaseParam.get_value`
        on the child. The parent's setter will be called when you
        :func:`~puzzlepiece.param.BaseParam.set_value` on a child.

        You may need to override this method when creating params that have a different call
        signature for ``__init__``. Additional arguments can then be provided with ``kwargs``.

        See :func:`puzzlepiece.piece.Popup.add_child_params` for a quick way of adding child
        params to a popup.

        :param kwargs: Additional arguments to pass when creating the child.
        """
        # Only make an explicit setter if this param has an explicit setter.
        # The other case is handled via a Signal below, once the child
        # param is created.
        setter = None if self._setter is None else (lambda value: self.set_value(value))

        # child params always have a getter, to make the direction of data flow clear.
        def getter():
            return self.get_value()

        kwargs = kwargs or {}

        child = type(self)(
            self._name,
            self._value,
            setter=setter,
            getter=getter,
            format=self._format,
            _type=self._type,
            **kwargs,
        )

        if self._setter is None:
            # If no explicit setter, just set the parent param whenever the child updates
            child.changed.connect(lambda: self.set_value(child.value))
        elif self._value is not None:
            # When a param is created and has an explicit setter, it will be highlighted
            # red to indicate the setter has not been called. Here we remove the highlight
            # for the child if the parent's setter has been called already.
            child.setAutoFillBackground(False)

        return child

    @property
    def type(self):
        """
        The fixed type of this param. The values set with
        :func:`puzzlepiece.param.BaseParam.set_value` will be cast to this type,
        and those returned by :func:`puzzlepiece.param.BaseParam.get_value`
        will be of this type.
        """
        return self._type

    @property
    def visible(self):
        """
        Bool flag indicating whether the param is displayed in the GUI.
        """
        return self._visible

    def keyPressEvent(self, event):
        """
        # Pressing enter triggers the set action.

        :meta private:
        """
        if (
            event.key() == QtCore.Qt.Key.Key_Enter
            or event.key() == QtCore.Qt.Key.Key_Return
        ):
            self.set_value()
            # Move focus out of the input, so other keyboard shortcuts can be processed
            self.setFocus()
        else:
            super().keyPressEvent(event)


class ParamInt(BaseParam):
    """
    A param with an integer input field. See the :func:`~puzzlepiece.param.spinbox` decorator below
    for how to use this in your Piece.
    """

    _type = int

    def __init__(
        self,
        name,
        value,
        v_min,
        v_max,
        setter,
        getter=None,
        visible=True,
        v_step=1,
        *args,
        **kwargs,
    ):
        self._v_min = v_min
        self._v_max = v_max
        self._v_step = v_step
        super().__init__(name, value, setter, getter, visible, *args, **kwargs)

    def _make_input(self, value=None, connect=None):
        """:meta private:"""
        input = QtWidgets.QSpinBox()
        input.setMinimum(self._v_min)
        input.setMaximum(self._v_max)
        input.setSingleStep(self._v_step)
        input.setGroupSeparatorShown(True)
        if value is not None:
            input.setValue(value)
        if connect is not None:
            input.valueChanged.connect(lambda x: connect())
        return input, True

    def _input_set_value(self, value):
        """:meta private:"""
        self.input.blockSignals(True)
        self.input.setValue(value)
        self.input.blockSignals(False)

    def _input_get_value(self):
        """:meta private:"""
        return self.input.value()

    def make_child_param(self, kwargs=None):
        return super().make_child_param(
            kwargs={
                "v_min": self._v_min,
                "v_max": self._v_max,
                "v_step": self._v_step,
            }
        )


class ParamFloat(ParamInt):
    """
    A param with a float input field. See the :func:`~puzzlepiece.param.spinbox` decorator below
    for how to use this in your Piece.
    """

    _type = float

    def _make_input(self, value=None, connect=None):
        """:meta private:"""
        input = QtWidgets.QDoubleSpinBox()
        input.setGroupSeparatorShown(True)
        input.setMinimum(self._v_min)
        input.setMaximum(self._v_max)
        input.setSingleStep(self._v_step)
        if value is not None:
            input.setValue(value)
        if connect is not None:
            input.valueChanged.connect(connect)
        return input, True


class _Slider(QtWidgets.QWidget):
    def __init__(self, value, v_min, v_max, v_step, format="{:.2f}"):
        self._v_min = v_min
        self._v_max = v_max
        self._v_step = v_step
        self._format = format
        super().__init__()

        layout = QtWidgets.QHBoxLayout()
        self.setLayout(layout)

        self.input = input = QtWidgets.QSlider(QtCore.Qt.Orientation.Horizontal)
        input.setMinimum(int(np.round(v_min / v_step)))
        input.setMaximum(int(np.round(v_max / v_step)))
        input.setTickPosition(input.TickPosition.TicksBelow)
        layout.addWidget(input)
        self.valueChanged = self.input.valueChanged
        self.input.valueChanged.connect(self._set_label)

        self.input_label = QtWidgets.QLabel()
        layout.addWidget(self.input_label)

        if value is not None:
            input.setValue(int(np.round(value) / v_step))
            self._set_label()

    def _set_label(self):
        value = self.input.value() * self._v_step
        self.input_label.setText(self._format.format(value))

    def setValue(self, value):
        self.input.setValue(int(np.round(value / self._v_step)))

    def value(self):
        return self.input.value() * self._v_step


class ParamSlider(ParamFloat):
    """
    A param with a slider, float or int depending on provided value and step.
    See the :func:`~puzzlepiece.param.slider` decorator below
    for how to use this in your Piece.
    """

    def _make_input(self, value=None, connect=None):
        if (
            isinstance(value, int)
            and isinstance(self._v_min, int)
            and isinstance(self._v_min, int)
            and isinstance(self._v_step, int)
        ):
            self._type = int

        format = "{:.2f}" if self._type is float else "{}"
        input = _Slider(value, self._v_min, self._v_max, self._v_step, format)
        if value is not None:
            input.setValue(value)
        if connect is not None:
            input.valueChanged.connect(connect)
        return input, True


class ParamText(BaseParam):
    """
    A param with a text input field. See the :func:`~puzzlepiece.param.text` decorator below
    for how to use this in your Piece.
    """

    _type = str

    def _make_input(self, value=None, connect=None):
        """:meta private:"""
        input = QtWidgets.QLineEdit()
        if value is not None:
            input.setText(value)
        if connect is not None:
            input.textChanged.connect(connect)
        return input, True

    def _input_set_value(self, value):
        """:meta private:"""
        self.input.blockSignals(True)
        self.input.setText(value)
        self.input.blockSignals(False)

    def _input_get_value(self):
        """:meta private:"""
        return self.input.text()


class ParamCheckbox(BaseParam):
    """
    A param with a checkbox input field. See the :func:`~puzzlepiece.param.checkbox` decorator below
    for how to use this in your Piece.

    This param runs :func:`puzzlepiece.param.Param.set_value` automatically when the checkbox
    is pressed, so an additional setter button is not displayed.
    """

    _type = int
    # TODO: handle exceptions better!

    def __init__(self, name, value, setter, getter=None, visible=True, *args, **kwargs):
        super().__init__(name, value, setter, getter, visible, *args, **kwargs)

    def _make_input(self, value=None, connect=None):
        """:meta private:"""
        input = QtWidgets.QCheckBox()
        if value is not None:
            input.setChecked(bool(value))
        # Handling the input change signal differently here to allow for some error handling
        # when the checkbox is clicked
        self._connected_click_handler = connect
        input.clicked.connect(self._click_handler)
        return input, False

    def _input_set_value(self, value):
        """:meta private:"""
        self.input.setChecked(bool(value))
        # if self._connected_click_handler is not None:
        #     self._connected_click_handler()

    def _input_get_value(self):
        """:meta private:"""
        return int(self.input.isChecked())

    def _click_handler(self, _):
        try:
            if self._connected_click_handler is not None:
                self._connected_click_handler()
            if self._setter is not None:
                # If there's a setter, we need to explicitly call set_value here
                self.set_value()
        except Exception as e:
            # Flip back the checkbox if the click resulted in an error
            self.input.setChecked(not (self.input.isChecked()))
            raise e


class _PartialAccessor:
    def __init__(self, param):
        self.param = param

    def __setitem__(self, key, value):
        self.param._value.__setitem__(key, value)
        self.param.set_value()


class ParamArray(BaseParam):
    """
    A param that stores a numpy array. There is no GUI input, the Param simply displays the
    dimensions of the array, and indicates when the data has been updated.

    The array can be modified through programmatic interaction with setter or getter functions
    (for example the array can be obtained from a hardware spectrometer), or treated as a variable
    and set using :func:`~puzzlepiece.param.BaseParam.set_value`.
    """

    _type = np.asarray

    def __init__(
        self,
        name,
        value,
        setter=None,
        getter=None,
        visible=True,
        _type=None,
        *args,
        **kwargs,
    ):
        self._indicator_state = True
        self._partial_accessor = _PartialAccessor(self)
        super().__init__(
            name, value, setter, getter, visible, _type=_type, *args, **kwargs
        )

    @property
    def set_partial(self):
        """
        Use this property to set values to slices of the stored numpy array, using
        any slicing methods that a numpy array accepts::

            puzzle['piece'].params['image'].set_partial[100:200, :] = 128

        This will call the param's setter if there's one, and in general
        acts like :func:`~puzzlepiece.param.BaseParam.set_value`.
        """
        return self._partial_accessor

    def _make_input(self, value=None, connect=None):
        """
        :meta private:
        """
        input = QtWidgets.QLabel()
        if value is not None:
            input.setText(self._format_array(value))
        return input, True

    def _input_set_value(self, value):
        """
        :meta private:
        """
        self.input.setText(self._format_array(value))

    def _input_get_value(self):
        """
        :meta private:
        """
        return self._value

    def set_value(self, value=None):
        """
        This method overrides :func:`puzzlepiece.param.BaseParam.set_value`.
        See there for documentation.

        :meta private:
        """
        # Small check to account for the label being set twice
        # in super().set_value(value) when the
        # value argument is None
        if value is not None:
            self._indicator_state = not self._indicator_state
        return super().set_value(value)

    def _format_array(self, value):
        self._indicator_state = not self._indicator_state
        return f"array{value.shape} {'◧' if self._indicator_state else '◨'}"


class ParamDropdown(BaseParam):
    """
    A param storing a string that also provides a dropdown. The user can edit the text field directly,
    and pressing enter will add the current value to the dropdown. A list of values can also be provided
    when creating this param, and they will populate the dropdown as soon as the app starts.

    The default param value provided here as `value` will either be selected from the dropdown if
    in `values`, or inserted into the text field as if a user typed it if not in `values`.

    :param values: List of default values available in the dropdown (can be None)
    """

    _type = str

    def __init__(
        self,
        name,
        value,
        values,
        setter=None,
        getter=None,
        visible=True,
        *args,
        **kwargs,
    ):
        if values is None:
            self._values = []
        else:
            self._values = list(values)
        super().__init__(name, value, setter, getter, visible, *args, **kwargs)

    def _make_input(self, value=None, connect=None):
        """:meta private:"""
        input = QtWidgets.QComboBox(editable=True)

        # Add the possible values
        input.addItems([str(x) for x in self._values])

        if value is not None:
            value = str(value)
            if index := input.findData(value) > -1:
                input.setCurrentIndex(index)
            else:
                input.setCurrentText(value)

        if connect is not None:
            input.currentTextChanged.connect(connect)

        return input, True

    def _input_set_value(self, value):
        """:meta private:"""
        value = str(value)
        self.input.blockSignals(True)

        if index := self.input.findData(value) > -1:
            self.input.setCurrentIndex(index)
        else:
            self.input.setCurrentText(value)

        self.input.blockSignals(False)

    def _input_get_value(self):
        """:meta private:"""
        return self.input.currentText()

    def make_child_param(self, kwargs=None):
        return super().make_child_param(
            kwargs={
                "values": self._values,
            }
        )


class ParamProgress(BaseParam):
    """
    A param with a progress bar. See the :func:`~puzzlepiece.param.progress` decorator below
    for how to use this in your Piece.
    """

    _type = float

    def _make_input(self, value=None, connect=None):
        """:meta private:"""
        input = QtWidgets.QProgressBar()
        input.setMinimum(0)
        input.setMaximum(1000)
        if value is not None:
            input.setValue(value)
        return input, True

    def _input_set_value(self, value):
        """:meta private:"""
        if value < 0:
            self.input.setMaximum(0)
        else:
            self.input.setMaximum(1000)
            self.input.setValue(int(value * 1000))

    def _input_get_value(self):
        """:meta private:"""
        return self.input.value()

    def iter(self, iterable):
        """
        Iterable wrapper that automatically updates the progress bar while the provided
        iterable is iterated over. This is quite similar to how the console progress bar
        ``tqdm`` works: https://tqdm.github.io/

        For example::

            for i in piece.params['progress'].iter(range(10)):
                print(i)
                # this will automatically update the progress bar as the iteration goes on
                puzzle.process_events() # you may need to refresh the Puzzle to display changes

        :param iterable: an iterable to wrap
        :rtype: iterable
        """
        if hasattr(iterable, "__len__"):
            length = len(iterable)
        else:
            length = -1

        for i, value in enumerate(iterable):
            self.set_value(i / length)
            yield value
        self.set_value(1)


def wrap_setter(piece, setter):
    """
    We wrap the setter function such that it can be called without passing
    a reference to the Piece as the first argument

    :meta private:
    """
    if setter is not None:

        @wraps(setter)
        def wrapper(value):
            return setter(piece, value)
    else:
        wrapper = None
    return wrapper


def wrap_getter(piece, getter):
    """
    We wrap the getter function such that it can be called without passing
    a reference to the Piece as the first argument

    :meta private:
    """
    if getter is not None:

        @wraps(getter)
        def wrapper():
            return getter(piece)
    else:
        wrapper = None
    return wrapper


# The decorator syntax in Python is a little confusing
def base_param(piece, name, value, visible=True, format="{}"):
    """
    A decorator generator for registering a :class:`~puzzlepiece.param.BaseParam` in a Piece's
    :func:`~puzzlepiece.piece.Piece.define_params` method with a given **setter**.

    This will simply display the current value with no option to edit it.

    To register a param and mark a function as its setter, do this::

        @puzzlepiece.param.base_param(self, 'param_name', 0)
        def param_setter(self, value):
            print(value)

    To register a param without a setter, call this function to get a decorator, and then pass ``None`` to that to indicate
    that a setter doesn't exist::

        puzzlepiece.param.base_param(self, 'param_name', 0)(None)

    Some of the decorators here decorate getters, and some decorate setters, depending on what is the more sensible default
    -- a :func:`~puzzlepiece.param.readout` decorates a getter, as it is meant to display obtained values, and a
    :func:`~puzzlepiece.param.spinbox` decorates a setter, as it's meant to be an easy way to input and set values. If you
    need to also register the other function, use the :func:`puzzlepiece.param.BaseParam.set_getter`
    and :func:`puzzlepiece.param.BaseParam.set_setter` decorators::

        @puzzlepiece.param.base_param(self, 'position', 0)
        def position(self, value):
            self.sdk.set_position(value)

        @position.set_getter(self)
        def position(self):
            return self.sdk.get_position()

    :param piece: The :class:`~puzzle.piece.Piece` this param should be registered with. Usually `self`, as this method should
      be called from within :func:`puzzlepiece.piece.Piece.define_params`
    :param name: a unique (per Piece) name for the param
    :param value: default display value for the param. If there is a setter, the setter will not be called, and the stored value
      will remain None until the param is explicitly set (either through :func:`~puzzlepiece.param.BaseParam.set_value`, or
      pressing the set button.)
    :param visible: bool flag, determined if a GUI component will be shown for this param.
    :param format: Default format for displaying the param if a custom input is not defined, and in
      :func:`puzzplepiece.parse.format`. For example `{:.2f}`, see https://pyformat.info/ for further details.

    :returns: A decorator that can be applied to a setter function. The decorator returns the param object when called.
    """

    # The base_param function *creates and returns a decorator*,
    # it's just the way to enable passing parameters when decorating something
    def decorator(setter):
        wrapper = wrap_setter(piece, setter)

        # Register the param with the Piece
        piece.params[name] = BaseParam(
            name, value, setter=wrapper, getter=None, visible=visible, format=format
        )

        # Return the newly created Param
        return piece.params[name]

    return decorator


def readout(piece, name, visible=True, format="{}", _type=None):
    """
    A decorator generator for registering a :class:`~puzzlepiece.param.BaseParam` in a Piece's
    :func:`~puzzlepiece.piece.Piece.define_params` method with a given **getter**.

    This will simply display the current value with no option to edit it.

    See :func:`~puzzlepiece.param.base_param` for more details.
    """

    def decorator(getter):
        wrapper = wrap_getter(piece, getter)
        piece.params[name] = BaseParam(
            name,
            None,
            setter=None,
            getter=wrapper,
            visible=visible,
            format=format,
            _type=_type,
        )
        return piece.params[name]

    return decorator


def spinbox(piece, name, value, v_min=-1e9, v_max=1e9, visible=True, v_step=1):
    """
    A decorator generator for registering a :class:`~puzzlepiece.param.ParamInt` or :class:`~puzzlepiece.param.ParamFloat`
    in a Piece's :func:`~puzzlepiece.piece.Piece.define_params` method with a given **setter**.

    The type of param registered depends on whether the ``value`` given is an int of a float.

    This will display a Qt spinbox - a numeric input box.

    See :func:`~puzzlepiece.param.base_param` for more details.

    :param v_min: The minimum value accepted by the spinbox.
    :param v_max: The maximum value accepted by the spinbox.
    :param v_step: The step change when spinbox buttons are used.
    """

    def decorator(setter):
        wrapper = wrap_setter(piece, setter)

        if isinstance(value, int):
            piece.params[name] = ParamInt(
                name,
                value,
                int(v_min),
                int(v_max),
                setter=wrapper,
                getter=None,
                visible=visible,
                v_step=int(v_step),
            )
        else:
            piece.params[name] = ParamFloat(
                name, value, v_min, v_max, wrapper, None, visible, v_step
            )
        return piece.params[name]

    return decorator


def slider(piece, name, value, v_min=0, v_max=1, visible=True, v_step=0.05):
    """
    A decorator generator for registering a :class:`~puzzlepiece.param.ParamSlider`
    in a Piece's :func:`~puzzlepiece.piece.Piece.define_params` method with a given **setter**.

    This will display a slider with the specified properties and a label.
    It will be an int or a float slider depending on the type of the values provided here.

    See :func:`~puzzlepiece.param.base_param` for more details.
    """

    def decorator(setter):
        wrapper = wrap_setter(piece, setter)
        piece.params[name] = ParamSlider(
            name, value, v_min, v_max, wrapper, None, visible, v_step
        )
        return piece.params[name]

    return decorator


def text(piece, name, value, visible=True):
    """
    A decorator generator for registering a :class:`~puzzlepiece.param.ParamText`
    in a Piece's :func:`~puzzlepiece.piece.Piece.define_params` method with a given **setter**.

    This will display a single line textbox.

    See :func:`~puzzlepiece.param.base_param` for more details.
    """

    def decorator(setter):
        wrapper = wrap_setter(piece, setter)
        piece.params[name] = ParamText(name, value, wrapper, None, visible)
        return piece.params[name]

    return decorator


def checkbox(piece, name, value, visible=True):
    """
    A decorator generator for registering a :class:`~puzzlepiece.param.ParamCheckbox`
    in a Piece's :func:`~puzzlepiece.piece.Piece.define_params` method with a given **setter**.

    This will display a checkbox as the input, and the param takes the values 0 and 1 for
    unchecked and checked states.

    This param runs puzzlepiece.param.Param.set_value() automatically when the checkbox is pressed,
    so an additional setter button is not displayed.

    See :func:`~puzzlepiece.param.base_param` for more details.
    """

    def decorator(setter):
        wrapper = wrap_setter(piece, setter)
        piece.params[name] = ParamCheckbox(name, value, wrapper, None, visible)
        return piece.params[name]

    return decorator


def array(piece, name, visible=True):
    """
    A decorator generator for registering a :class:`~puzzlepiece.param.ParamArray`
    in a Piece's :func:`~puzzlepiece.piece.Piece.define_params` method with a given **getter**.

    This will display the shape of the stored array with no option to edit it and an indicator
    showing when the value changes.

    See :func:`~puzzlepiece.param.base_param` for more details.
    """

    def decorator(getter):
        wrapper = wrap_getter(piece, getter)
        piece.params[name] = ParamArray(
            name, None, setter=None, getter=wrapper, visible=visible
        )
        return piece.params[name]

    return decorator


def dropdown(piece, name, value, visible=True):
    """
    A decorator generator for registering a :class:`~puzzlepiece.param.ParamDropdown`
    in a Piece's :func:`~puzzlepiece.piece.Piece.define_params`.

    It should decorate a function that returns a list of default values to populate the dropdown,
    for example::

        @puzzlepiece.param.dropdown(self, 'param_name', '')
        def param_values(self, value):
            return self.sdk.discover_devices()

    It can also be used with a set list of defaults, or with no defaults at all::

        puzzlepiece.param.dropdown(self, 'param_name', 'one')(['one', 'two', 'three'])
        puzzlepiece.param.dropdown(self, 'param_name', 'four')(None)

    Setters and getters can then be added using :func:`puzzlepiece.param.BaseParam.set_getter`
    and :func:`puzzlepiece.param.BaseParam.set_setter` decorators::

        @puzzlepiece.param.dropdown(self, 'serial_number', '')
        def serial_number(self, value):
            return self.sdk.discover_devices()

        @serial_number.set_getter(self)
        def serial_number(self):
            return self.sdk.get_serial()

        @serial_number.set_setter(self)
        def serial_number(self, value):
            return self.sdk.set_serial(value)

    The returned param displays a dropdown and stores a string. The user can edit the dropdown's
    text field directly, and pressing enter will add the current value to the dropdown.

    The default param value provided here as `value` will either be selected from the dropdown if
    available, or inserted into the text field as if a user typed it otherwise.
    """

    def decorator(values):
        if callable(values):
            # `values` can be a function that returns a list of values
            values = values(piece)
        piece.params[name] = ParamDropdown(name, value, values, None, None, visible)
        return piece.params[name]

    return decorator


def progress(piece, name, visible=True):
    """
    A decorator generator for registering a :class:`~puzzlepiece.param.ParamProgress` in a Piece's
    :func:`~puzzlepiece.piece.Piece.define_params` method with a given **getter**.

    This will display the current progress value on a scale of 0 to 1 with no option to edit it.

    See :func:`~puzzlepiece.param.base_param` for more details.
    """

    def decorator(getter):
        wrapper = wrap_getter(piece, getter)
        piece.params[name] = ParamProgress(
            name, None, setter=None, getter=wrapper, visible=visible
        )
        return piece.params[name]

    return decorator



import puzzlepiece as pzp
from pyqtgraph.Qt import QtWidgets
import numpy as np
import time


class Piece(pzp.Piece):
    def __init__(self, puzzle):
        super().__init__(puzzle)

    def define_params(self):
        pzp.param.text(self, "params", "laser:power")(None)
        pzp.param.text(self, "action", "run:lightfield:Acquire")(None)
        pzp.param.text(self, "break", "lightfield:saturated")(None)

        pzp.param.spinbox(self, "start", 8)(None)
        pzp.param.spinbox(self, "end", 60)(None)
        pzp.param.spinbox(self, "step", 2)(None)
        pzp.param.spinbox(self, "finish", 0)(None)

    def param_layout(self, wrap=1):
        return super().param_layout(wrap)

    def define_actions(self):
        @pzp.action.define(self, "Scan")
        def scan(self):
            values = np.arange(
                self.params["start"].get_value(),
                self.params["end"].get_value(),
                self.params["step"].get_value(),
            )
            params = pzp.parse.parse_params(
                self.params["params"].get_value(), self.puzzle
            )
            break_param = (
                pzp.parse.parse_params(self.params["break"].get_value(), self.puzzle)[0]
                if len(self.params["break"].get_value())
                else None
            )
            command = self.params["action"].get_value()
            self.progress_bar.setMaximum(len(values))
            self.stop = False

            for i, value in enumerate(values):
                for param in params:
                    param.set_value(value)
                time.sleep(0.05)
                pzp.parse.run(command, self.puzzle)
                self.progress_bar.setValue(i + 1)
                self.puzzle.process_events()
                if self.stop or (break_param is not None and break_param.get_value()):
                    break
            for param in params:
                param.set_value(self.params["finish"].get_value())
            # Maybe plot it?

    def custom_layout(self):
        layout = QtWidgets.QVBoxLayout()

        self.progress_bar = QtWidgets.QProgressBar()
        self.progress_bar.setMaximum(1)
        self.progress_bar.setValue(0)
        layout.addWidget(self.progress_bar)

        return layout



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "puzzlepiece"
copyright = "2024, Jakub Dranczewski"
author = "Jakub Dranczewski"
release = "0.1"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc", "sphinx.ext.viewcode", "sphinx.ext.napoleon",
    "nbsphinx", "sphinx.ext.autosectionlabel"]

templates_path = ["_templates"]
exclude_patterns = []

autodoc_member_order = "bysource"

# autodoc_mock_imports = ["pyqtgraph", "QtPy", "numpy"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



import puzzlepiece as pzp
from pyqtgraph.Qt import QtWidgets


class Piece(pzp.Piece):
    def __init__(self, puzzle):
        super().__init__(puzzle)
        self.stop = False

    def define_params(self):
        pzp.param.text(self, "iterator", "")(None)
        pzp.param.text(self, "pre", "")(None)
        pzp.param.text(self, "post", "")(None)
        pzp.param.spinbox(self, "counter", 0, visible=False)(None)

    def define_actions(self):
        @pzp.action.define(self, "Save")
        def save(self):
            fname = str(QtWidgets.QFileDialog.getSaveFileName(self, "Save file...")[0])
            with open(fname, "w") as f:
                f.write("#pre " + self.params["pre"].get_value() + "\n")
                f.write("#post " + self.params["post"].get_value() + "\n")
                f.write(self.text.toPlainText())

        @pzp.action.define(self, "Open")
        def _open(self):
            fname = str(QtWidgets.QFileDialog.getOpenFileName(self, "Open file...")[0])
            with open(fname, "r") as f:
                text = ""
                for line in f.readlines():
                    if "#pre " in line:
                        self.params["pre"].set_value(line[5:-1])
                    elif "#post " in line:
                        self.params["post"].set_value(line[6:-1])
                    else:
                        text += line
                self.text.setPlainText(text)

        @pzp.action.define(self, "Run")
        def run(self):
            self.stop = False
            iter_name = self.params["iterator"].get_value()

            try:
                pzp.parse.run(self.params["pre"].get_value(), self.puzzle)

                if len(iter_name):
                    iterator = self.puzzle.pieces[iter_name].iterator()
                    for step in iterator:
                        pzp.parse.run(self.text.toPlainText(), self.puzzle)
                        if self.stop:
                            self.stop = False
                            return
                else:
                    pzp.parse.run(self.text.toPlainText(), self.puzzle)
            finally:
                pzp.parse.run(self.params["post"].get_value(), self.puzzle)

    def custom_layout(self):
        layout = QtWidgets.QVBoxLayout()

        self.text = QtWidgets.QPlainTextEdit()
        self.text.setPlainText("")
        layout.addWidget(self.text)

        return layout



from pyqtgraph.Qt import QtCore
from functools import wraps


class Action(QtCore.QObject):
    """
    An action is a function that a :class:`~puzzlepiece.piece.Piece` object can call.

    It will be given a button in the GUI (if ``visible`` is True), and can be executed from code
    in other Pieces or elsewhere as if it was a method::

        puzzle['piece_name'].actions['action_name']()

    Any arguments provided when calling an action will be passed to the registered function.

    To register an action, use the :func:`~puzzlepiece.action.define` decorator as shown below.

    :param function: The function to call when the action is executed.
    :param parent: The Piece this action belongs to.
    :param shortcut: A keyboard shortcut for this action, works only when the Piece is visible.
    :param visible: Bool flag, whether a button for the action is generated in the GUI.
    """

    #: A Qt signal emitted when the action is executed.
    called = QtCore.Signal()

    def __init__(self, function, parent, shortcut=None, visible=True):
        self.function = function
        self.parent = parent
        #: Keyboard shortcut associated with the param.
        self.shortcut = shortcut
        # See https://doc.qt.io/qt-6/qt.html#Key-enum for acceptable values
        self._visible = visible
        super().__init__()

    def __call__(self, *args, **kwargs):
        # Bring the Piece into view if in a folder
        self.parent.elevate()
        self.function(*args, **kwargs)
        self.called.emit()

    @property
    def visible(self):
        """
        Bool flag, indicates whether this action is visible as a button in the GUI.
        """
        return self._visible

    def make_child_action(self):
        """
        Create and return a child action that calls the same callable.

        See :func:`puzzlepiece.piece.Popup.add_child_actions` for a quick way of adding child
        actions to a popup.
        """
        child = Action(self.function, self.parent, self.shortcut)
        return child


def define(piece, name, shortcut=None, visible=True):
    """
    A decorator generator for registering a :class:`~puzzlepiece.action.Action` in a Piece's
    :func:`~puzzlepiece.piece.Piece.define_action` method with a given function.

    To register an action for a function, do this::

        @puzzlepiece.action.define(self, 'action_name')
        def action(self):
            print("Hello world!")

    :param piece: The :class:`~puzzle.piece.Piece` this param should be registered with. Usually `self`, as this method should
      be called from within :func:`puzzlepiece.piece.Piece.define_actions`
    :param name: a unique (per Piece) name for the action
    :param shortcut: The keyboard shortcut for this action. See https://doc.qt.io/qt-6/qt.html#Key-enum for possible values.
      Example: ``QtCore.Qt.Key.Key_F1``
    :param visible: bool flag, determined if a GUI button will be shown for this param.
    """

    def decorator(action):
        @wraps(action)
        def wrapper(*args, **kwargs):
            return action(piece, *args, **kwargs)

        action_object = Action(wrapper, piece, shortcut, visible)
        piece.actions[name] = action_object
        if shortcut:
            piece.shortcuts[shortcut] = action_object
        return action_object

    return decorator



import puzzlepiece as pzp
from puzzlepiece.pieces import random_number, plotter, scan_value, script


def main():
    # Define the containing app
    app = pzp.QApp([])

    # Create a Puzzle
    puzzle = pzp.Puzzle(app, "Basic example", debug=True)

    # Add Pieces
    puzzle.add_piece("random", random_number.Piece, 0, 0)
    folder = puzzle.add_folder(0, 1)
    folder.add_piece("plotter", plotter.Piece)
    folder.add_piece("scan_value", scan_value.Piece)
    folder.add_piece("script", script.Piece)

    # Run a setup script
    puzzle.run("""
set:plotter:param:random:number
set:scan_value:params:random:max
set:scan_value:obtain:random:number
set:scan_value:end:100
set:scan_value:finish:10
""")

    # Show the app
    puzzle.show()
    app.exec()


if __name__ == "__main__":
    main()



from . import piece
from . import puzzle
from . import param
from . import readout
from . import action
from . import parse
from . import threads

Piece = piece.Piece
Puzzle = puzzle.Puzzle
QApp = puzzle.QApp

__all__ = [piece, puzzle, param, readout, action, parse, threads, Piece, Puzzle, QApp]



import puzzlepiece as pzp
import random


class Piece(pzp.Piece):
    # We define a Piece object to contain the random number generator

    # We give it 'params' within this function (these will appear as inputs/value displays)
    def define_params(self):
        # Some params don't need setters, they're just variables that impact other things.
        # In that case we pass None to the defining decorator.
        pzp.param.spinbox(self, "min", 0)(None)
        pzp.param.spinbox(self, "max", 10)(None)

        # Some params have a 'setter' function, which sets a value, like a laser's power.
        # In that case we make the function (which takes a value) and decorate it with
        # a param-defining decorator
        @pzp.param.spinbox(self, "seed", 0)
        def seed(self, value):
            random.seed(value)

        # Some params have a 'getter' function, which returns a value, like a powermeter's reading
        # In that case we make the function (which returns a value) and decorate it with
        # a readout-param-defining decorator
        @pzp.param.readout(self, "number")
        def random_number(self):
            return random.randint(
                self.params["min"].get_value(), self.params["max"].get_value()
            )

    # We give it 'actions' within this function (these will appear as buttons)
    def define_actions(self):
        # Sometimes an action is needed, like homing a moving stage
        # In that case we make the function (which performs the action) and decorate it with
        # an action-defining decorator
        @pzp.action.define(self, "Dialog")
        def print_something(self):
            pzp.parse.run(
                "prompt:In a range between {random:min} and {random:max}, your number is {random:number}",
                self.puzzle,
            )



import puzzlepiece as pzp
from pyqtgraph.Qt import QtWidgets
import pyqtgraph as pg
import time
import numpy as np


class Piece(pzp.Piece):
    def __init__(self, puzzle):
        super().__init__(puzzle, custom_horizontal=True)
        self.running = False
        self.times = []
        self.data = []

    def define_params(self):
        pzp.param.text(self, "param", "plotter:max")(None)
        pzp.param.spinbox(self, "max", 100)(None)

        @pzp.param.spinbox(self, "sleep", 0.1)
        def set_sleep(self, value):
            self.timer.sleep = value

    def define_actions(self):
        @pzp.action.define(self, "Clear")
        def clear(self):
            self.times = []
            self.data = []

    def add_point(self, value):
        self.data.append(value)
        self.times.append(time.time())

        max_l = self.params["max"].get_value()
        self.times = self.times[-max_l:]
        self.data = self.data[-max_l:]

        if len(self.data) > 1:
            td = np.array(self.times) - self.times[0]
            if np.amax(td) > 60:
                td /= 60
            self.plot_line.setData(td, self.data)

    def get_value(self):
        # The task that happens in the side-thread should be simple and not touch the Widgets or
        # the main thread too much - it can take a while though, like a lengthy data acquisition
        param = pzp.parse.parse_params(self.params["param"].get_value(), self.puzzle)[0]
        return param.get_value()

    def custom_layout(self):
        layout = QtWidgets.QVBoxLayout()

        # The thread runs self.get_value repeatedly, which returns a value...
        self.timer = pzp.threads.PuzzleTimer(
            "Live", self.puzzle, self.get_value, self.params["sleep"].get_value()
        )
        # ... this value is then passed to self.add_point through a Signal
        self.timer.returned.connect(self.add_point)

        layout.addWidget(self.timer)
        self.params["sleep"].set_value()  # Set the sleep value to the default one

        self.pw = pg.PlotWidget()
        layout.addWidget(self.pw)
        self.plot = self.pw.getPlotItem()
        self.plot_line = self.plot.plot([0], [0], symbol="o", symbolSize=3)

        return layout

    def call_stop(self):
        self.timer.stop()



import puzzlepiece as pzp
from pyqtgraph.Qt import QtWidgets


class TParamPiece(pzp.Piece):
    def define_params(self):
        pzp.param.base_param(self, "base_param", 0)(None)

        self._setter_param_value = None

        @pzp.param.base_param(self, "setter_param", 1)
        def setter_param(self, value):
            self._setter_param_value = value

        @pzp.param.base_param(self, "setter_return_param", 0)
        def setter_return_param(self, value):
            return value + 1

        @pzp.param.readout(self, "getter_param", _type=int)
        def getter_param(self):
            return 1

        self._setter_getter_param_value = 0

        @pzp.param.base_param(self, "setter_getter_param", 0)
        def setter_getter_param(self, value):
            self._setter_getter_param_value = value

        @setter_getter_param.set_getter(self)
        def setter_getter_param_getter(self):
            return self._setter_getter_param_value + 1

        self._setter_getter_return_param_value = 0

        @pzp.param.base_param(self, "setter_getter_return_param", 0)
        def setter_getter_return_param(self, value):
            self._setter_getter_return_param_value = value
            return value

        @setter_getter_return_param.set_getter(self)
        def setter_getter_return_param_getter(self):
            return self._setter_getter_return_param_value + 1

        @pzp.param.base_param(self, "error_param", 0)
        def error_param(self, value):
            # return 0
            raise Exception("Setter exception")

        @error_param.set_getter(self)
        def error_param_getter(self):
            # return 0
            raise Exception("Getter exception")

        @pzp.param.checkbox(self, "error_checkbox", 0)
        def error_param(self, value):
            raise Exception("Setter exception")

        pzp.param.base_param(self, "float_param", 0.0)(None)
        pzp.param.base_param(self, "format_param", 0.0, format="{:.2f}")(None)
        pzp.param.spinbox(self, "input_param", 0.0)(None)
        pzp.param.text(self, "text_param", "")(None)


def test_base_param(qtbot, qapp):
    puzzle = pzp.Puzzle(qapp, "Test params")
    puzzle.add_piece("test", TParamPiece(puzzle), 0, 0)
    puzzle.show()

    count = [0]

    def count_calls(count=count):
        count[0] += 1

    puzzle["test"].params["base_param"].changed.connect(count_calls)

    assert count[0] == 0
    assert puzzle["test"].params["base_param"].value == 0
    assert puzzle["test"].params["base_param"].get_value() == 0
    assert puzzle["test"].params["base_param"].input.text() == "0"
    assert count[0] == 0

    puzzle["test"].params["base_param"].set_value(1)
    assert count[0] == 1
    assert puzzle["test"].params["base_param"].value == 1
    assert puzzle["test"].params["base_param"].get_value() == 1
    assert puzzle["test"].params["base_param"].input.text() == "1"


def test_setter_param(qtbot, qapp):
    puzzle = pzp.Puzzle(qapp, "Test params")
    puzzle.add_piece("test", TParamPiece(puzzle), 0, 0)
    puzzle.show()

    count = [0]

    def count_calls(count=count):
        count[0] += 1

    puzzle["test"].params["setter_param"].changed.connect(count_calls)

    assert count[0] == 0
    assert puzzle["test"]._setter_param_value is None
    assert puzzle["test"].params["setter_param"].value is None
    assert puzzle["test"].params["setter_param"].get_value() is None
    assert puzzle["test"].params["setter_param"].input.text() == "1"
    assert count[0] == 0

    puzzle["test"].params["setter_param"].set_value()
    assert count[0] == 1
    assert puzzle["test"]._setter_param_value == 1
    assert puzzle["test"].params["setter_param"].value == 1
    assert puzzle["test"].params["setter_param"].get_value() == 1
    assert puzzle["test"].params["setter_param"].input.text() == "1"

    puzzle["test"].params["setter_param"].set_value(2)
    assert count[0] == 2
    assert puzzle["test"]._setter_param_value == 2
    assert puzzle["test"].params["setter_param"].value == 2
    assert puzzle["test"].params["setter_param"].get_value() == 2
    assert puzzle["test"].params["setter_param"].input.text() == "2"

    puzzle["test"].params["setter_param"]._input_set_value(3)
    assert count[0] == 2
    assert puzzle["test"].params["setter_param"].input.text() == "3"
    puzzle["test"].params["setter_param"]._set_button.click()
    assert count[0] == 3
    assert puzzle["test"]._setter_param_value == 3
    assert puzzle["test"].params["setter_param"].value == 3
    assert puzzle["test"].params["setter_param"].get_value() == 3
    assert puzzle["test"].params["setter_param"].input.text() == "3"


def test_setter_return_param(qtbot, qapp):
    puzzle = pzp.Puzzle(qapp, "Test params")
    puzzle.add_piece("test", TParamPiece(puzzle), 0, 0)
    puzzle.show()

    count = [0]

    def count_calls(count=count):
        count[0] += 1

    puzzle["test"].params["setter_return_param"].changed.connect(count_calls)

    assert count[0] == 0
    assert puzzle["test"].params["setter_return_param"].value is None
    assert puzzle["test"].params["setter_return_param"].get_value() is None
    assert puzzle["test"].params["setter_return_param"].input.text() == "0"
    assert count[0] == 0

    puzzle["test"].params["setter_return_param"].set_value()
    assert count[0] == 1
    assert puzzle["test"].params["setter_return_param"].value == 1
    assert puzzle["test"].params["setter_return_param"].get_value() == 1
    assert puzzle["test"].params["setter_return_param"].input.text() == "1"

    puzzle["test"].params["setter_return_param"].set_value(2)
    assert count[0] == 2
    assert puzzle["test"].params["setter_return_param"].value == 3
    assert puzzle["test"].params["setter_return_param"].get_value() == 3
    assert puzzle["test"].params["setter_return_param"].input.text() == "3"

    puzzle["test"].params["setter_return_param"]._input_set_value(4)
    assert count[0] == 2
    assert puzzle["test"].params["setter_return_param"].input.text() == "4"
    puzzle["test"].params["setter_return_param"]._set_button.click()
    assert count[0] == 3
    assert puzzle["test"].params["setter_return_param"].value == 5
    assert puzzle["test"].params["setter_return_param"].get_value() == 5
    assert puzzle["test"].params["setter_return_param"].input.text() == "5"


def test_getter_param(qtbot, qapp):
    puzzle = pzp.Puzzle(qapp, "Test params")
    puzzle.add_piece("test", TParamPiece(puzzle), 0, 0)
    puzzle.show()

    count = [0]

    def count_calls(count=count):
        count[0] += 1

    puzzle["test"].params["getter_param"].changed.connect(count_calls)

    assert count[0] == 0
    assert puzzle["test"].params["getter_param"].value is None
    assert puzzle["test"].params["getter_param"].input.text() == ""
    assert count[0] == 0

    puzzle["test"].params["getter_param"].get_value()
    assert count[0] == 1
    assert puzzle["test"].params["getter_param"].value == 1
    assert puzzle["test"].params["getter_param"].input.text() == "1"

    puzzle["test"].params["getter_param"].get_value()
    assert count[0] == 2
    assert puzzle["test"].params["getter_param"].value == 1
    assert puzzle["test"].params["getter_param"].input.text() == "1"

    # Updated design: set_value in general emits the signal,
    # even if the input field is unchanged
    puzzle["test"].params["getter_param"].set_value()
    assert count[0] == 3
    assert puzzle["test"].params["getter_param"].value == 1
    assert puzzle["test"].params["getter_param"].input.text() == "1"

    # In contrast, changing the value of the input with the hidden function _input_set_value
    # should in general not emit the signal
    puzzle["test"].params["getter_param"]._input_set_value(2)
    assert count[0] == 3
    assert puzzle["test"].params["getter_param"].value == 1
    assert puzzle["test"].params["getter_param"].input.text() == "2"

    puzzle["test"].params["getter_param"].set_value(3)
    assert count[0] == 4
    assert puzzle["test"].params["getter_param"].value == 3
    assert puzzle["test"].params["getter_param"].input.text() == "3"

    puzzle["test"].params["getter_param"].get_value()
    assert count[0] == 5
    assert puzzle["test"].params["getter_param"].value == 1
    assert puzzle["test"].params["getter_param"].input.text() == "1"


def test_precision_param(qtbot, qapp):
    puzzle = pzp.Puzzle(qapp, "Test params")
    puzzle.add_piece("test", TParamPiece(puzzle), 0, 0)
    puzzle.show()

    # The base_param has an int format, so it should cast values given to int
    puzzle["test"].params["base_param"].set_value(0.1234)
    assert puzzle["test"].params["base_param"].value == 0
    assert puzzle["test"].params["base_param"].get_value() == 0

    # The float_param has a float format, so it should retain precision
    puzzle["test"].params["float_param"].set_value(0.1234)
    assert puzzle["test"].params["float_param"].value == 0.1234
    assert puzzle["test"].params["float_param"].get_value() == 0.1234

    # how about one with a format that has fewer decimal points?
    puzzle["test"].params["format_param"].set_value(0.1234)
    assert puzzle["test"].params["format_param"].value == 0.1234
    assert puzzle["test"].params["format_param"].get_value() == 0.1234

    # how about one with an input?
    puzzle["test"].params["input_param"].set_value(0.1234)
    assert puzzle["test"].params["input_param"].value == 0.1234
    assert puzzle["test"].params["input_param"].get_value() == 0.1234


def test_text_param(qtbot, qapp):
    puzzle = pzp.Puzzle(qapp, "Test params")
    puzzle.add_piece("test", TParamPiece(puzzle), 0, 0)
    puzzle.show()

    count = [0]

    def count_calls(count=count):
        count[0] += 1

    puzzle["test"].params["text_param"].changed.connect(count_calls)

    assert count[0] == 0

    # The internal method _input_set_value should not result in the Signal being emitted
    # or the internal value changing
    puzzle["test"].params["text_param"]._input_set_value("A")
    assert count[0] == 0
    assert puzzle["test"].params["text_param"].value == ""
    assert puzzle["test"].params["text_param"].get_value() == ""

    # In contrast, changing the text field directly (like the user typing)
    # should change the internal value and emit the Signal
    puzzle["test"].params["text_param"].input.setText("B")
    assert count[0] == 1
    assert puzzle["test"].params["text_param"].value == "B"
    assert puzzle["test"].params["text_param"].get_value() == "B"

    # same for changing through the set_value method, which should also update the text field
    puzzle["test"].params["text_param"].set_value("C")
    assert count[0] == 2
    assert puzzle["test"].params["text_param"].value == "C"
    assert puzzle["test"].params["text_param"].get_value() == "C"
    assert puzzle["test"].params["text_param"]._input_get_value() == "C"


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    puzzle = pzp.Puzzle(app, "Test params")
    puzzle.add_piece("test", TParamPiece(puzzle), 0, 0)
    from puzzlepiece.pieces import script

    puzzle.add_piece("script", script.Piece(puzzle), 0, 1)
    puzzle.show()
    app.exec()



import puzzlepiece as pzp
import pyqtgraph as pg
from pyqtgraph.Qt import QtWidgets
import numpy as np
import time


class Piece(pzp.Piece):
    def __init__(self, puzzle):
        super().__init__(puzzle, custom_horizontal=True)
        self.x = []
        self.y = []
        self.stop = False

    def define_params(self):
        pzp.param.text(self, "params", "")(None)
        pzp.param.text(self, "obtain", "")(None)

        pzp.param.spinbox(self, "start", 0.0)(None)
        pzp.param.spinbox(self, "end", 11.0)(None)
        pzp.param.spinbox(self, "step", 1.0)(None)
        pzp.param.spinbox(self, "finish", 0.0)(None)
        pzp.param.text(self, "filename", "")(None)

    def param_layout(self, wrap=1):
        return super().param_layout(wrap)

    def define_actions(self):
        @pzp.action.define(self, "Scan")
        def scan(self):
            values = np.arange(
                self.params["start"].get_value(),
                self.params["end"].get_value(),
                self.params["step"].get_value(),
            )
            params = pzp.parse.parse_params(
                self.params["params"].get_value(), self.puzzle
            )
            obtain = pzp.parse.parse_params(
                self.params["obtain"].get_value(), self.puzzle
            )[0]
            self.progress_bar.setMaximum(len(values))

            self.x = []
            self.y = []
            self.stop = False
            for i, value in enumerate(values):
                for param in params:
                    param.set_value(value)
                self.x.append(value)
                time.sleep(0.05)
                self.y.append(obtain.get_value())
                self.progress_bar.setValue(i + 1)
                self.plot_line.setData(self.x, self.y)
                self.puzzle.process_events()

                if self.stop:
                    break
            for param in params:
                param.set_value(self.params["finish"].get_value())

        @pzp.action.define(self, "Save")
        def save(self):
            out = np.stack((self.x, self.y)).T
            filename = self.params["filename"].get_value()
            filename = pzp.parse.format(filename, self.puzzle)
            np.savetxt(filename, out, delimiter=",")

        @pzp.action.define(self, "Name")
        def choose_file(self):
            fname = str(QtWidgets.QFileDialog.getSaveFileName(self, "Save file...")[0])
            self.params["filename"].set_value(fname)

    def custom_layout(self):
        layout = QtWidgets.QVBoxLayout()

        self.progress_bar = QtWidgets.QProgressBar()
        self.progress_bar.setMaximum(1)
        self.progress_bar.setValue(0)
        layout.addWidget(self.progress_bar)

        self.pw = pg.PlotWidget()
        layout.addWidget(self.pw)
        self.plot = self.pw.getPlotItem()
        self.plot_line = self.plot.plot([0], [0], symbol="o", symbolSize=3)

        return layout
